import io
import re
import zipfile
from typing import List, Optional

import wx
from lxml import etree

import trelby.config as config
import trelby.gutil as gutil
import trelby.misc as misc
import trelby.screenplay as screenplay
import trelby.titles as titles
import trelby.util as util
from trelby.importdlg import ImportDlg
from trelby.indent import Indent
from trelby.line import Line

# special linetype that means that indent contains action and scene lines,
# and scene lines are the ones that begin with "EXT." or "INT."
SCENE_ACTION = -2

# special linetype that means don't import those lines; useful for page
# numbers etc
IGNORE = -3


# like importTextFile, but for Adobe Story files.
def importAstx(fileName, frame):
    # astx files are xml files. The textlines can be found under
    # AdobeStory/document/stream/section/scene/paragraph which contain
    # one or more textRun/break elements, to be joined. The paragraph
    # attribute "element" gives us the element style.

    data = util.loadFile(fileName, frame, -1)

    if data == None:
        return None

    if len(data) == 0:
        wx.MessageBox("File is empty.", "Error", wx.OK, frame)

        return None

    elemMap = {
        "Action": screenplay.ACTION,
        "Character": screenplay.CHARACTER,
        "Dialog": screenplay.DIALOGUE,
        "Parenthetical": screenplay.PAREN,
        "SceneHeading": screenplay.SCENE,
        "Shot": screenplay.SHOT,
        "Transition": screenplay.TRANSITION,
    }

    try:
        root = etree.XML(data)
    except etree.XMLSyntaxError as e:
        wx.MessageBox("Error parsing file: %s" % e, "Error", wx.OK, frame)
        return None

    lines = []

    def addElem(eleType, items):
        # if elem ends in a newline, last line is empty and useless;
        # get rid of it
        if not items[-1] and (len(items) > 1):
            items = items[:-1]

        for s in items[:-1]:
            lines.append(Line(screenplay.LB_FORCED, eleType, util.cleanInput(s)))

        lines.append(Line(screenplay.LB_LAST, eleType, util.cleanInput(items[-1])))

    for para in root.xpath("/AdobeStory/document/stream/section/scene/paragraph"):
        lt = elemMap.get(para.get("element"), screenplay.ACTION)

        items = []
        s = ""

        for text in para:
            if text.tag == "textRun" and text.text:
                s += text.text
            elif text.tag == "break":
                items.append(s.rstrip())
                s = ""

        items.append(s.rstrip())

        addElem(lt, items)

    if not lines:
        wx.MessageBox("File has no content.", "Error", wx.OK, frame)
        return None

    return lines


# like importTextFile, but for fadein files.
def importFadein(fileName, frame):
    # Fadein file is a zipped document.xml file.
    # the .xml is in open screenplay format:
    # http://sourceforge.net/projects/openscrfmt/files/latest/download

    try:
        with zipfile.ZipFile(fileName) as fadein_file:
            with fadein_file.open("document.xml") as fadein_document:
                data = fadein_document.read()
    except:
        wx.MessageBox("File is not a valid .fadein file.", "Error", wx.OK, frame)
        return None

    if data == None:
        return None

    if len(data) == 0:
        wx.MessageBox("File is empty.", "Error", wx.OK, frame)
        return None

    elemMap = {
        "Action": screenplay.ACTION,
        "Character": screenplay.CHARACTER,
        "Dialogue": screenplay.DIALOGUE,
        "Parenthetical": screenplay.PAREN,
        "Scene Heading": screenplay.SCENE,
        "Shot": screenplay.SHOT,
        "Transition": screenplay.TRANSITION,
    }

    try:
        root = etree.XML(data)
    except etree.XMLSyntaxError as e:
        wx.MessageBox("Error parsing file: %s" % e, "Error", wx.OK, frame)
        return None

    lines = []

    def addElem(eleType, lns):
        # if elem ends in a newline, last line is empty and useless;
        # get rid of it
        if not lns[-1] and (len(lns) > 1):
            lns = lns[:-1]

        for s in lns[:-1]:
            lines.append(Line(screenplay.LB_FORCED, eleType, util.cleanInput(s)))

        lines.append(Line(screenplay.LB_LAST, eleType, util.cleanInput(lns[-1])))

    # removes html formatting from s, and returns list of lines.
    # if s is None, return a list with single empty string.
    re_rem = [r"<font[^>]*>", r"<size[^>]*>", r"<bgcolor[^>]*>"]
    rem = [
        "<b>",
        "</b>",
        "<i>",
        "</i>",
        "<u>",
        "</u>",
        "</font>",
        "</size>",
        "</bgcolor>",
    ]

    def sanitizeStr(s):
        if s:
            s = "" + s
            for r in re_rem:
                s = re.sub(r, "", s)
            for r in rem:
                s = s.replace(r, "")

            if s:
                return s.split("<br>")
            else:
                return [""]
        else:
            return [""]

    for para in root.xpath("paragraphs/para"):
        # check for notes/synopsis, import as Note.
        if para.get("note"):
            lt = screenplay.NOTE
            items = sanitizeStr("" + para.get("note"))
            addElem(lt, items)

        if para.get("synopsis"):
            lt = screenplay.NOTE
            items = sanitizeStr("" + para.get("synopsis"))
            addElem(lt, items)

        # look for the <style> and <text> tags. Bail if no <text> found.
        styl = para.xpath("style")
        txt = para.xpath("text")
        if txt:
            if styl:
                lt = elemMap.get(styl[0].get("basestylename"), screenplay.ACTION)
            else:
                lt = screenplay.ACTION

            items = sanitizeStr(txt[0].text)

            if (lt == screenplay.PAREN) and items and (items[0][0] != "("):
                items[0] = "(" + items[0]
                items[-1] = items[-1] + ")"
        else:
            continue

        addElem(lt, items)

    if len(lines) == 0:
        wx.MessageBox("The file contains no importable lines", "Error", wx.OK, frame)
        return None

    return lines


# like importTextFile, but for Celtx files.
def importCeltx(fileName, frame):
    # Celtx files are zipfiles, and the script content is within a file
    # called "script-xxx.html", where xxx can be random.
    try:
        z = zipfile.ZipFile(fileName)
    except:
        wx.MessageBox("File is not a valid Celtx script file.", "Error", wx.OK, frame)
        return None

    files = z.namelist()
    scripts = [s for s in files if s.startswith("script")]

    if len(scripts) == 0:
        wx.MessageBox(
            "Unable to find script in this Celtx file.", "Error", wx.OK, frame
        )
        return None

    f = z.open(scripts[0])
    content = f.read()
    z.close()

    if not content:
        wx.MessageBox("Script seems to be empty.", "Error", wx.OK, frame)
        return None

    elemMap = {
        "action": screenplay.ACTION,
        "character": screenplay.CHARACTER,
        "dialog": screenplay.DIALOGUE,
        "parenthetical": screenplay.PAREN,
        "sceneheading": screenplay.SCENE,
        "shot": screenplay.SHOT,
        "transition": screenplay.TRANSITION,
        "act": screenplay.ACTBREAK,
    }

    try:
        parser = etree.HTMLParser()
        root = etree.XML(content, parser)
    except etree.XMLSyntaxError as e:
        wx.MessageBox("Error parsing file: %s" % e, "Error", wx.OK, frame)
        return None

    lines = []

    def addElem(eleType, lns):
        # if elem ends in a newline, last line is empty and useless;
        # get rid of it
        if not lns[-1] and (len(lns) > 1):
            lns = lns[:-1]

        for s in lns[:-1]:
            lines.append(Line(screenplay.LB_FORCED, eleType, util.cleanInput(s)))

        lines.append(Line(screenplay.LB_LAST, eleType, util.cleanInput(lns[-1])))

    for para in root.xpath("/html/body/p"):
        items = []
        for line in para.itertext():
            text = line.replace("\n", " ").strip()
            if text != "":
                items.append(text)

        lt = elemMap.get(para.get("class"), screenplay.ACTION)

        if items:
            addElem(lt, items)

    if len(lines) == 0:
        wx.MessageBox("The file contains no importable lines", "Error", wx.OK, frame)
        return None

    return lines


# like importTextFile, but for Final Draft files.
def importFDX(fileName, frame):
    elemMap = {
        "Action": screenplay.ACTION,
        "Character": screenplay.CHARACTER,
        "Dialogue": screenplay.DIALOGUE,
        "Parenthetical": screenplay.PAREN,
        "Scene Heading": screenplay.SCENE,
        "Shot": screenplay.SHOT,
        "Transition": screenplay.TRANSITION,
    }

    data = util.loadFile(fileName, frame, -1)

    if data == None:
        return None

    if len(data) == 0:
        wx.MessageBox("File is empty.", "Error", wx.OK, frame)

        return None

    # https://github.com/trelby/trelby/issues/61
    data = data.replace("\u2028", "\n")

    try:
        root = etree.XML(data.encode("UTF-8"))
        lines = []

        def addElem(eleType, eleText):
            lns = eleText.split("\n")

            # if elem ends in a newline, last line is empty and useless;
            # get rid of it
            if not lns[-1] and (len(lns) > 1):
                lns = lns[:-1]

            for s in lns[:-1]:
                lines.append(Line(screenplay.LB_FORCED, eleType, util.cleanInput(s)))

            lines.append(Line(screenplay.LB_LAST, eleType, util.cleanInput(lns[-1])))

        for para in root.xpath("Content//Paragraph"):
            addedNote = False
            et = para.get("Type")

            # Check for script notes
            s = ""
            for notes in para.xpath("ScriptNote/Paragraph/Text"):
                if notes.text:
                    s += notes.text

                # FD has AdornmentStyle set to "0" on notes with newline.
                if notes.get("AdornmentStyle") == "0":
                    s += "\n"

            if s:
                addElem(screenplay.NOTE, s)
                addedNote = True

            # "General" has embedded Dual Dialogue paragraphs inside it;
            # nothing to do for the General element itself.
            #
            # If no type is defined (like inside scriptnote), skip.
            if (et == "General") or (et is None):
                continue

            s = ""
            for text in para.xpath("Text"):
                # text.text is None for paragraphs with no text, and +=
                # blows up trying to add a string object and None, so
                # guard against that
                if text.text:
                    s += text.text

            # don't remove paragraphs with no text, unless that paragraph
            # contained a scriptnote
            if s or not addedNote:
                lt = elemMap.get(et, screenplay.ACTION)
                addElem(lt, s)

        if len(lines) == 0:
            wx.MessageBox(
                "The file contains no importable lines", "Error", wx.OK, frame
            )
            return None

        return lines

    except etree.XMLSyntaxError as e:
        wx.MessageBox("Error parsing file: %s" % e, "Error", wx.OK, frame)
        return None


# import Fountain files.
# http://fountain.io
def importFountain(fileName, frame, titlePages):
    # regular expressions for fountain markdown.
    # https://github.com/vilcans/screenplain/blob/master/screenplain/richstring.py
    ire = re.compile(
        # one star
        r"\*"
        # anything but a space, then text
        r"([^\s].*?)"
        # finishing with one star
        r"\*"
        # must not be followed by star
        r"(?!\*)"
    )
    bre = re.compile(
        # two stars
        r"\*\*"
        # must not be followed by space
        r"(?=\S)"
        # inside text
        r"(.+?[*_]*)"
        # finishing with two stars
        r"(?<=\S)\*\*"
    )
    ure = re.compile(
        # underline
        r"_"
        # must not be followed by space
        r"(?=\S)"
        # inside text
        r"([^_]+)"
        # finishing with underline
        r"(?<=\S)_"
    )
    boneyard_re = re.compile("/\\*.*?\\*/", flags=re.DOTALL)

    # random magicstring used to escape literal star '\*'
    literalstar = "Aq7RR"

    # returns s with markdown formatting removed.
    def unmarkdown(s):
        s = s.replace("\\*", literalstar)
        for style in (bre, ire, ure):
            s = style.sub(r"\1", s)
        return s.replace(literalstar, "*")

    data = util.loadFile(fileName, frame, -1)

    if data == None:
        return None

    if len(data) == 0:
        wx.MessageBox("File is empty.", "Error", wx.OK, frame)
        return None

    inf = []
    inf.append(misc.CheckBoxItem("Import titles to title page."))
    inf.append(misc.CheckBoxItem("Import titles as action lines.", selected=False))
    inf.append(misc.CheckBoxItem("Remove unsupported formatting markup."))
    inf.append(misc.CheckBoxItem("Import section/synopsis as notes."))

    dlg = misc.CheckBoxDlg(
        frame, "Fountain import options", inf, "Import options:", False
    )

    if dlg.ShowModal() != wx.ID_OK:
        dlg.Destroy()
        return None, titlePages

    importTitlePage = inf[0].selected
    importTitles = inf[1].selected
    removeMarkdown = inf[2].selected
    importSectSyn = inf[3].selected

    # pre-process data - fix newlines, remove boneyard.
    data = util.fixNL(data)
    data = boneyard_re.sub("", data)
    prelines = data.split("\n")
    for i in range(len(prelines)):
        try:
            util.toLatin1(prelines[i])
        except:
            prelines[i] = util.cleanInput("" + prelines[i].decode("UTF-8", "ignore"))
    lines = []

    tabWidth = 4
    lns = []
    TWOSPACE = "  "
    skipone = False

    # First check if title lines are present:
    c = 0
    while c < len(prelines):
        if prelines[c] != "":
            c = c + 1
        else:
            break

    # prelines[0:i] are the first bunch of lines, that could be titles.
    # Our check for title is simple:
    #   - the line does not start with 'fade'
    #   - the first line has a single ':'

    if c > 0:
        l = util.toInputStr(prelines[0].expandtabs(tabWidth).lstrip().lower())
        if not l.startswith("fade") and l.count(":") == 1:
            # these are title lines. Now do what the user requested.
            if importTitlePage:
                # Extract usable title page lines
                kwPattern = re.compile(r"^([A-za-z][^:]+:)")
                continuePattern = re.compile(
                    r"^\s\s\s+"
                )  # line begins with 3 or more spaces
                titleIdx = 0
                gatheredTitles = []
                rejects = []
                l = prelines[0]
                while l != "":
                    match = kwPattern.split(l)
                    match = [ll.strip() for ll in match]
                    if len(match) == 3:  # keyword line
                        if match[2] == "":
                            follows = []
                            titleIdx += 1  # gather indented prelines
                            nextline = continuePattern.match(prelines[titleIdx])
                            while nextline:
                                follows.append(prelines[titleIdx].strip("_* \t"))
                                titleIdx += 1
                                nextline = continuePattern.match(prelines[titleIdx])
                            titleIdx -= 1
                            gatheredTitles.append([match[1], follows])
                        else:
                            gatheredTitles.append([match[1], [match[2].strip("_* \t")]])
                    else:
                        l += TWOSPACE
                        rejects.append(l)
                    titleIdx += 1
                    l = prelines[titleIdx]
                # replace default titles with imported ones
                leftX = 15.0
                leftY = 220.0
                centreX = 0.0
                centreY = 150.0
                for t in gatheredTitles:
                    titlenr = -1
                    if t[0] == "Title:":
                        titlenr = 0
                    elif t[0] == "Author:" or t[0] == "Authors:":
                        titlenr = 1
                        t[1] = ["by", ""] + t[1]
                    elif t[0] == "Contact:":
                        titlenr = 2
                    else:
                        titlenr = len(titlePages[0])
                        if t[0] == "Credit:" or t[0] == "Source:":
                            centred = True
                            thisX = centreX
                            thisY = centreY
                            centreY += 5
                        else:
                            centred = False
                            thisX = leftX
                            thisY = leftY
                            leftY -= 5
                        if len(t[1]) == 1:
                            newItems = [t[0] + " " + t[1][0]]
                        else:
                            newItems = [t[0]] + t[1]
                        newTitleString = titles.TitleString(
                            newItems, y=thisY, isCentered=centred
                        )
                        titlePages[0].append(newTitleString)
                        continue
                    targetTitle = titlePages[0][titlenr]
                    targetTitle.items = t[1]
                # user might request that titles go into both title page & script
                if not importTitles:
                    rejects += prelines[c + 1 :]
                    prelines = rejects
            if importTitles:
                # add TWOSPACE to all the title lines.
                for i in range(c):
                    prelines[i] += TWOSPACE
            elif not importTitlePage:
                # remove these lines
                prelines = prelines[c + 1 :]

    for l in prelines:
        if l != TWOSPACE:
            lines.append(util.toInputStr(l.expandtabs(tabWidth)))
        else:
            lines.append(TWOSPACE)

    linesLen = len(lines)

    def isPrevEmpty():
        if lns and lns[-1].text == "":
            return True
        return False

    def isPrevType(ltype):
        return lns and lns[-1].lt == ltype

    # looks ahead to check if next line is not empty
    def isNextEmpty(i):
        return i + 1 < len(lines) and lines[i + 1] == ""

    def getPrevType():
        if lns:
            return lns[-1].lt
        else:
            return screenplay.ACTION

    def isParen(s):
        return s.startswith("(") and s.endswith(")")

    def isScene(s):
        if s.endswith(TWOSPACE):
            return False
        if s.startswith(".") and not s.startswith(".."):
            return True
        tmp = s.upper()
        if (
            re.match(r"^(INT|EXT|EST)[ .]", tmp)
            or re.match(r"^(INT\.?/EXT\.?)[ .]", tmp)
            or re.match(r"^I/E[ .]", tmp)
        ):
            return True
        return False

    def isTransition(s):
        return (s.isupper() and s.endswith("TO:")) or (
            s.startswith(">") and not s.endswith("<")
        )

    def isCentered(s):
        return s.startswith(">") and s.endswith("<")

    def isPageBreak(s):
        return s.startswith("===") and s.lstrip("=") == ""

    def isNote(s):
        return s.startswith("[[") and s.endswith("]]")

    def isSection(s):
        return s.startswith("#")

    def isSynopsis(s):
        return s.startswith("=") and not s.startswith("==")

    # first pass - identify linetypes
    for i in range(linesLen):
        if skipone:
            skipone = False
            continue

        s = lines[i]
        sl = s.lstrip()
        # mark as ACTION by default.
        line = Line(screenplay.LB_FORCED, screenplay.ACTION, s)

        # Start testing lines for element type. Go in order:
        # Scene Character, Paren, Dialog, Transition, Note.

        if s == "" or isCentered(s) or isPageBreak(s):
            # do nothing - import as action.
            pass

        elif s == TWOSPACE:
            line.lt = getPrevType()

        elif isScene(s):
            line.lt = screenplay.SCENE
            if sl.startswith("."):
                line.text = sl[1:]
            else:
                line.text = sl

        elif isTransition(sl) and isPrevEmpty() and isNextEmpty(i):
            line.lt = screenplay.TRANSITION
            if line.text.startswith(">"):
                line.text = sl[1:].lstrip()

        elif s.isupper() and isPrevEmpty() and not isNextEmpty(i):
            line.lt = screenplay.CHARACTER
            if s.endswith(TWOSPACE):
                line.lt = screenplay.ACTION

        elif isParen(sl) and (
            isPrevType(screenplay.CHARACTER) or isPrevType(screenplay.DIALOGUE)
        ):
            line.lt = screenplay.PAREN

        elif (
            isPrevType(screenplay.CHARACTER)
            or isPrevType(screenplay.DIALOGUE)
            or isPrevType(screenplay.PAREN)
        ):
            line.lt = screenplay.DIALOGUE

        elif isNote(sl):
            line.lt = screenplay.NOTE
            line.text = sl.strip("[]")

        elif isSection(s) or isSynopsis(s):
            if not importSectSyn:
                if isNextEmpty(i):
                    skipone = True
                continue

            line.lt = screenplay.NOTE
            line.text = sl.lstrip("=#")

        if line.text == TWOSPACE:
            pass

        elif line.lt != screenplay.ACTION:
            line.text = line.text.lstrip()

        else:
            tmp = line.text.rstrip()
            # we don't support center align, so simply add required indent.
            if isCentered(tmp):
                tmp = tmp[1:-1].strip()
                width = frame.panel.ctrl.sp.cfg.getType(screenplay.ACTION).width
                if len(tmp) < width:
                    tmp = " " * ((width - len(tmp)) // 2) + tmp
            line.text = tmp

        if removeMarkdown:
            line.text = unmarkdown(line.text)
            if line.lt == screenplay.CHARACTER and line.text.endswith("^"):
                line.text = line.text[:-1]

        lns.append(line)

    ret = []

    # second pass helper functions.
    def isLastLBForced():
        return ret and ret[-1].lb == screenplay.LB_FORCED

    def makeLastLBLast():
        if ret:
            ret[-1].lb = screenplay.LB_LAST

    def isRetPrevType(t):
        return ret and ret[-1].lt == t

    # second pass - remove unneeded empty lines, and fix the linebreaks.
    for ln in lns:
        if ln.text == "":
            if isLastLBForced():
                makeLastLBLast()
            else:
                ret.append(ln)

        elif not isRetPrevType(ln.lt):
            makeLastLBLast()
            ret.append(ln)

        else:
            ret.append(ln)

    makeLastLBLast()
    return ret, titlePages


# import text file from fileName, return list of Line objects for the
# screenplay or None if something went wrong. returned list always
# contains at least one line.
def importTextFile(fileName: str, frame: wx.Frame) -> Optional[List[Line]]:

    data = util.loadFile(fileName, frame, -1)

    if data == None:
        return None

    if len(data) == 0:
        wx.MessageBox("File is empty.", "Error", wx.OK, frame)

        return None

    data = util.fixNL(data)
    lines = data.split("\n")

    tabWidth = 4

    # key = indent level, value = Indent
    indDict = {}

    for i in range(len(lines)):
        s = util.toInputStr(lines[i].rstrip().expandtabs(tabWidth))

        # don't count empty lines towards indentation statistics
        if s.strip() == "":
            lines[i] = ""

            continue

        cnt = util.countInitial(s, " ")

        ind = indDict.get(cnt)
        if not ind:
            ind = Indent(cnt)
            indDict[cnt] = ind

        tmp = s.upper()

        if util.multiFind(tmp, ["EXT.", "INT."]):
            ind.sceneStart += 1

        if util.multiFind(tmp, ["CUT TO:", "DISSOLVE TO:"]):
            ind.trans += 1

        if re.match(r"^ +\(.*\)$", tmp):
            ind.paren += 1

        ind.lines.append(s.lstrip())
        lines[i] = s

    if len(indDict) == 0:
        wx.MessageBox("File contains only empty lines.", "Error", wx.OK, frame)

        return None

    # scene/action indent
    setType(SCENE_ACTION, indDict, lambda v: v.sceneStart)

    # indent with most lines is dialogue in non-pure-action scripts
    setType(screenplay.DIALOGUE, indDict, lambda v: len(v.lines))

    # remaining indent with lines is character most likely
    setType(screenplay.CHARACTER, indDict, lambda v: len(v.lines))

    # transitions
    setType(screenplay.TRANSITION, indDict, lambda v: v.trans)

    # parentheticals
    setType(screenplay.PAREN, indDict, lambda v: v.paren)

    # some text files have this type of parens:
    #
    #        JOE
    #      (smiling and
    #       hopping along)
    #
    # this handles them.
    parenIndent = findIndent(indDict, lambda v: v.lt == screenplay.PAREN)
    if parenIndent != -1:
        paren2Indent = findIndent(
            indDict, lambda v, var: (v.lt == -1) and (v.indent == var), parenIndent + 1
        )

        if paren2Indent != -1:
            indDict[paren2Indent].lt = screenplay.PAREN

    # set line type to ACTION for any indents not recognized
    for v in indDict.values():
        if v.lt == -1:
            v.lt = screenplay.ACTION

    dlg = ImportDlg(frame, list(indDict.values()), SCENE_ACTION, IGNORE)

    if dlg.ShowModal() != wx.ID_OK:
        dlg.Destroy()

        return None

    dlg.Destroy()

    ret = []

    for i in range(len(lines)):
        s = lines[i]
        cnt = util.countInitial(s, " ")
        s = s.lstrip()
        sUp = s.upper()

        if s:
            lt = indDict[cnt].lt

            if lt == IGNORE:
                continue

            if lt == SCENE_ACTION:
                if s.startswith("EXT.") or s.startswith("INT."):
                    lt = screenplay.SCENE
                else:
                    lt = screenplay.ACTION

            if ret and (ret[-1].lt != lt):
                ret[-1].lb = screenplay.LB_LAST

            if lt == screenplay.CHARACTER:
                if sUp.endswith("(CONT'D)"):
                    s = sUp[:-8].rstrip()

            elif lt == screenplay.PAREN:
                if s == "(continuing)":
                    s = ""

            if s:
                line = screenplay.Line(screenplay.LB_SPACE, lt, s)
                ret.append(line)

        elif ret:
            ret[-1].lb = screenplay.LB_LAST

    if len(ret) == 0:
        ret.append(screenplay.Line(screenplay.LB_LAST, screenplay.ACTION))

    # make sure the last line ends an element
    ret[-1].lb = screenplay.LB_LAST

    return ret


# go through indents, find the one with maximum value in something, and
# set its linetype to given lt.
def setType(lt, indDict, func):
    maxCount = 0
    found = -1

    for v in indDict.values():
        # don't touch indents already set
        if v.lt != -1:
            continue

        val = func(v)

        if val > maxCount:
            maxCount = val
            found = v.indent

    if found != -1:
        indDict[found].lt = lt


# go through indents calling func(it, *vars) on each. return indent count
# for the indent func returns True, or -1 if it returns False for each.
def findIndent(indDict, func, *vars):
    for v in indDict.values():
        if func(v, *vars):
            return v.indent

    return -1



import functools
import operator

import trelby.misc as misc
import trelby.pdf as pdf
import trelby.pml as pml
import trelby.util as util
from trelby.pagelist import PageList


class LocationReport:
    # sr = SceneReport
    def __init__(self, sr):
        self.sp = sr.sp

        # key = scene name, value = LocationInfo. note that multiple keys
        # can point to the same LocationInfo.
        locations = {}

        # like locations, but this one stores per-scene information
        self.scenes = {}

        # make grouped scenes point to the same LocationInfos.
        for sceneList in self.sp.locations.locations:
            li = LocationInfo(self.sp)

            for scene in sceneList:
                locations[scene] = li

        # merge scene information for locations and store scene
        # information
        for si in sr.scenes:
            locations.setdefault(si.name, LocationInfo(self.sp)).addScene(si)

            self.scenes.setdefault(si.name, LocationInfo(self.sp)).addScene(si)

        # remove empty LocationInfos, sort them and store to a list
        tmp = []
        for li in locations.values():
            if (len(li.scenes) > 0) and (li not in tmp):
                tmp.append(li)

        def cmpfunc(a, b):
            return (a > b) - (a < b)

        def sortFunc(o1, o2):
            ret = cmpfunc(o2.lines, o1.lines)

            if ret != 0:
                return ret
            else:
                return cmpfunc(o1.scenes[0], o2.scenes[0])

        tmp = sorted(tmp, key=functools.cmp_to_key(sortFunc))

        self.locations = tmp

        # information about what to include (and yes, the comma is needed
        # to unpack the list)
        (self.INF_SPEAKERS,) = list(range(1))
        self.inf = []
        for s in ["Speakers"]:
            self.inf.append(misc.CheckBoxItem(s))

    def generate(self) -> bytes:
        tf = pml.TextFormatter(
            self.sp.cfg.paperWidth, self.sp.cfg.paperHeight, 15.0, 12
        )

        scriptLines = sum([li.lines for li in self.locations])

        for li in self.locations:
            tf.addSpace(5.0)

            # list of (scenename, lines_in_scene) tuples, which we sort in
            # DESC(lines_in_scene) ASC(scenename) order.
            tmp = [(scene, self.scenes[scene].lines) for scene in li.scenes]

            tmp.sort(key=operator.itemgetter(0))
            tmp.sort(key=operator.itemgetter(1), reverse=True)

            for scene, lines in tmp:
                if len(tmp) > 1:
                    pct = " (%d%%)" % util.pct(lines, li.lines)
                else:
                    pct = ""

                tf.addText("%s%s" % (scene, pct), style=pml.BOLD)

            tf.addSpace(1.0)

            tf.addWrappedText(
                "Lines: %d (%d%% action, %d%% of script),"
                " Scenes: %d, Pages: %d (%s)"
                % (
                    li.lines,
                    util.pct(li.actionLines, li.lines),
                    util.pct(li.lines, scriptLines),
                    li.sceneCount,
                    len(li.pages),
                    li.pages,
                ),
                "  ",
            )

            if self.inf[self.INF_SPEAKERS].selected:
                tf.addSpace(2.5)

                for it in util.sortDict(li.chars):
                    tf.addText("     %3d  %s" % (it[1], it[0]))

        return pdf.generate(tf.doc)


# information about one location
class LocationInfo:
    def __init__(self, sp):
        # number of scenes
        self.sceneCount = 0

        # scene names, e.g. ["INT. MOTEL ROOM - NIGHT", "EXT. MOTEL -
        # NIGHT"]
        self.scenes = []

        # total lines, excluding scene lines
        self.lines = 0

        # action lines
        self.actionLines = 0

        # page numbers
        self.pages = PageList(sp.getPageNumbers())

        # key = character name (upper cased), value = number of dialogue
        # lines
        self.chars = {}

    # add a scene. si = SceneInfo
    def addScene(self, si):
        if si.name not in self.scenes:
            self.scenes.append(si.name)

        self.sceneCount += 1
        self.lines += si.lines
        self.actionLines += si.actionLines
        self.pages += si.pages

        for name, dlines in si.chars.items():
            self.chars[name] = self.chars.get(name, 0) + dlines



# information about one indent level in imported text files.
class Indent:
    def __init__(self, indent):

        # indent level, i.e. spaces at the beginning
        self.indent = indent

        # lines with this indent, leading spaces removed
        self.lines = []

        # assigned line type, or -1 if not assigned yet.
        self.lt = -1

        # how many of the lines start with "EXT." or "INT."
        self.sceneStart = 0

        # how many of the lines have "CUT TO:" or "DISSOLVE TO:"
        self.trans = 0

        # how many of the lines have a form of "^ +\(.*)$", i.e. are most
        # likely parentheticals
        self.paren = 0



import copy

import wx

import trelby.gutil as gutil
import trelby.misc as misc
import trelby.pdf as pdf
import trelby.pml as pml
import trelby.titles as titles
import trelby.util as util
from trelby.titlespreview import TitlesPreview


class TitlesDlg(wx.Dialog):
    def __init__(self, parent, titles, cfg, cfgGl):
        wx.Dialog.__init__(
            self, parent, -1, "Title pages", style=wx.DEFAULT_DIALOG_STYLE
        )

        self.titles = titles
        self.cfg = cfg
        self.cfgGl = cfgGl

        # whether some events are blocked
        self.block = False

        self.setPage(0)

        vsizer = wx.BoxSizer(wx.VERTICAL)

        self.pageLabel = wx.StaticText(self, -1, "")
        vsizer.Add(self.pageLabel, 0, wx.ADJUST_MINSIZE)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        tmp = wx.Button(self, -1, "Add")
        hsizer.Add(tmp)
        self.Bind(wx.EVT_BUTTON, self.OnAddPage, id=tmp.GetId())
        gutil.btnDblClick(tmp, self.OnAddPage)

        self.delPageBtn = wx.Button(self, -1, "Delete")
        hsizer.Add(self.delPageBtn, 0, wx.LEFT, 10)
        self.Bind(wx.EVT_BUTTON, self.OnDeletePage, id=self.delPageBtn.GetId())
        gutil.btnDblClick(self.delPageBtn, self.OnDeletePage)

        self.moveBtn = wx.Button(self, -1, "Move")
        hsizer.Add(self.moveBtn, 0, wx.LEFT, 10)
        self.Bind(wx.EVT_BUTTON, self.OnMovePage, id=self.moveBtn.GetId())
        gutil.btnDblClick(self.moveBtn, self.OnMovePage)

        self.nextBtn = wx.Button(self, -1, "Next")
        hsizer.Add(self.nextBtn, 0, wx.LEFT, 10)
        self.Bind(wx.EVT_BUTTON, self.OnNextPage, id=self.nextBtn.GetId())
        gutil.btnDblClick(self.nextBtn, self.OnNextPage)

        vsizer.Add(hsizer, 0, wx.TOP, 5)

        vsizer.Add(wx.StaticLine(self, -1), 0, wx.EXPAND | wx.TOP | wx.BOTTOM, 10)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        vsizer2 = wx.BoxSizer(wx.VERTICAL)

        tmp = wx.StaticText(self, -1, "Strings:")
        vsizer2.Add(tmp)

        self.stringsLb = wx.ListBox(self, -1, size=(200, 150))
        vsizer2.Add(self.stringsLb)

        hsizer2 = wx.BoxSizer(wx.HORIZONTAL)

        self.addBtn = gutil.createStockButton(self, "Add")
        hsizer2.Add(self.addBtn)
        self.Bind(wx.EVT_BUTTON, self.OnAddString, id=self.addBtn.GetId())
        gutil.btnDblClick(self.addBtn, self.OnAddString)

        self.delBtn = gutil.createStockButton(self, "Delete")
        hsizer2.Add(self.delBtn, 0, wx.LEFT, 10)
        self.Bind(wx.EVT_BUTTON, self.OnDeleteString, id=self.delBtn.GetId())
        gutil.btnDblClick(self.delBtn, self.OnDeleteString)

        vsizer2.Add(hsizer2, 0, wx.TOP, 5)

        hsizer.Add(vsizer2)

        self.previewCtrl = TitlesPreview(self, self, self.cfg)
        util.setWH(self.previewCtrl, 150, 150)
        hsizer.Add(self.previewCtrl, 1, wx.EXPAND | wx.LEFT, 10)

        vsizer.Add(hsizer, 0, wx.EXPAND)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        hsizer.Add(wx.StaticText(self, -1, "Text:"), 0, wx.ALIGN_CENTER_VERTICAL)
        self.textEntry = wx.TextCtrl(
            self, -1, style=wx.TE_MULTILINE | wx.TE_DONTWRAP, size=(200, 75)
        )
        hsizer.Add(self.textEntry, 1, wx.LEFT, 10)
        self.Bind(wx.EVT_TEXT, self.OnMisc, id=self.textEntry.GetId())

        vsizer.Add(hsizer, 0, wx.EXPAND | wx.TOP, 20)

        # TODO: should use FlexGridSizer, like headersdlg, to get neater
        # layout

        hsizerTop = wx.BoxSizer(wx.HORIZONTAL)

        vsizer2 = wx.BoxSizer(wx.VERTICAL)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        hsizer.Add(wx.StaticText(self, -1, "Alignment:"), 0, wx.ALIGN_CENTER_VERTICAL)
        self.alignCombo = wx.ComboBox(self, -1, style=wx.CB_READONLY)

        for it in [
            ("Left", util.ALIGN_LEFT),
            ("Center", util.ALIGN_CENTER),
            ("Right", util.ALIGN_RIGHT),
        ]:
            self.alignCombo.Append(it[0], it[1])

        hsizer.Add(self.alignCombo, 0, wx.LEFT, 10)
        self.Bind(wx.EVT_COMBOBOX, self.OnMisc, id=self.alignCombo.GetId())

        vsizer2.Add(hsizer, 0, wx.TOP, 5)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        hsizer.Add(
            wx.StaticText(self, -1, "X / Y Pos (mm):"), 0, wx.ALIGN_CENTER_VERTICAL
        )
        self.xEntry = wx.TextCtrl(self, -1)
        hsizer.Add(self.xEntry, 0, wx.LEFT, 10)
        self.Bind(wx.EVT_TEXT, self.OnMisc, id=self.xEntry.GetId())
        self.yEntry = wx.TextCtrl(self, -1)
        hsizer.Add(self.yEntry, 0, wx.LEFT, 10)
        self.Bind(wx.EVT_TEXT, self.OnMisc, id=self.yEntry.GetId())

        vsizer2.Add(hsizer, 0, wx.TOP, 5)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        hsizer.Add(wx.StaticText(self, -1, "Font / Size:"), 0, wx.ALIGN_CENTER_VERTICAL)
        self.fontCombo = wx.ComboBox(self, -1, style=wx.CB_READONLY)

        for it in [
            ("Courier", pml.COURIER),
            ("Helvetica", pml.HELVETICA),
            ("Times-Roman", pml.TIMES_ROMAN),
        ]:
            self.fontCombo.Append(it[0], it[1])

        hsizer.Add(self.fontCombo, 0, wx.LEFT, 10)
        self.Bind(wx.EVT_COMBOBOX, self.OnMisc, id=self.fontCombo.GetId())

        self.sizeEntry = wx.SpinCtrl(self, -1, size=(50, -1))
        self.sizeEntry.SetRange(4, 288)
        self.Bind(wx.EVT_SPINCTRL, self.OnMisc, id=self.sizeEntry.GetId())
        self.sizeEntry.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)
        hsizer.Add(self.sizeEntry, 0, wx.LEFT, 10)

        vsizer2.Add(hsizer, 0, wx.TOP, 10)

        hsizerTop.Add(vsizer2)

        bsizer = wx.StaticBoxSizer(wx.StaticBox(self, -1, "Style"), wx.HORIZONTAL)

        vsizer2 = wx.BoxSizer(wx.VERTICAL)

        # wxGTK adds way more space by default than wxMSW between the
        # items, have to adjust for that
        pad = 0
        if misc.isWindows:
            pad = 5

        self.addCheckBox("Bold", self, vsizer2, pad)
        self.addCheckBox("Italic", self, vsizer2, pad)
        self.addCheckBox("Underlined", self, vsizer2, pad)

        bsizer.Add(vsizer2)

        hsizerTop.Add(bsizer, 0, wx.LEFT, 20)

        vsizer.Add(hsizerTop, 0, wx.TOP, 10)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        hsizer.Add((1, 1), 1)

        self.previewBtn = gutil.createStockButton(self, "Preview")
        hsizer.Add(self.previewBtn)

        cancelBtn = gutil.createStockButton(self, "Cancel")
        hsizer.Add(cancelBtn, 0, wx.LEFT, 10)

        okBtn = gutil.createStockButton(self, "OK")
        hsizer.Add(okBtn, 0, wx.LEFT, 10)

        vsizer.Add(hsizer, 0, wx.EXPAND | wx.TOP, 20)

        util.finishWindow(self, vsizer)

        self.Bind(wx.EVT_BUTTON, self.OnPreview, id=self.previewBtn.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnCancel, id=cancelBtn.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnOK, id=okBtn.GetId())

        self.Bind(wx.EVT_LISTBOX, self.OnStringsLb, id=self.stringsLb.GetId())

        # list of widgets that are specific to editing the selected string
        self.widList = [
            self.textEntry,
            self.xEntry,
            self.alignCombo,
            self.yEntry,
            self.fontCombo,
            self.sizeEntry,
            self.boldCb,
            self.italicCb,
            self.underlinedCb,
        ]

        self.updateGui()

        self.textEntry.SetFocus()

    def addCheckBox(self, name, parent, sizer, pad):
        cb = wx.CheckBox(parent, -1, name)
        self.Bind(wx.EVT_CHECKBOX, self.OnMisc, id=cb.GetId())
        sizer.Add(cb, 0, wx.TOP, pad)
        setattr(self, name.lower() + "Cb", cb)

    def OnOK(self, event):
        self.titles.sort()
        self.EndModal(wx.ID_OK)

    def OnCancel(self, event):
        self.EndModal(wx.ID_CANCEL)

    def OnPreview(self, event):
        doc = pml.Document(self.cfg.paperWidth, self.cfg.paperHeight)

        self.titles.generatePages(doc)
        tmp = pdf.generate(doc)
        gutil.showTempPDF(tmp, self.cfgGl, self)

    # set given page. 'page' can be an invalid value.
    def setPage(self, page):
        # selected page index or -1
        self.pageIndex = -1

        if self.titles.pages:
            self.pageIndex = 0

            if (page >= 0) and (len(self.titles.pages) > page):
                self.pageIndex = page

        # selected string index or -1
        self.tsIndex = -1

        if self.pageIndex == -1:
            return

        if len(self.titles.pages[self.pageIndex]) > 0:
            self.tsIndex = 0

    def OnKillFocus(self, event):
        self.OnMisc()

        # if we don't call this, the spin entry on wxGTK gets stuck in
        # some weird state
        event.Skip()

    def OnStringsLb(self, event=None):
        self.tsIndex = self.stringsLb.GetSelection()
        self.updateStringGui()

    def OnAddPage(self, event):
        self.titles.pages.append([])
        self.setPage(len(self.titles.pages) - 1)

        self.updateGui()

    def OnDeletePage(self, event):
        del self.titles.pages[self.pageIndex]
        self.setPage(0)

        self.updateGui()

    def OnMovePage(self, event):
        newIndex = (self.pageIndex + 1) % len(self.titles.pages)

        self.titles.pages[self.pageIndex], self.titles.pages[newIndex] = (
            self.titles.pages[newIndex],
            self.titles.pages[self.pageIndex],
        )

        self.setPage(newIndex)

        self.updateGui()

    def OnNextPage(self, event):
        self.setPage((self.pageIndex + 1) % len(self.titles.pages))

        self.updateGui()

    def OnAddString(self, event):
        if self.pageIndex == -1:
            return

        if self.tsIndex != -1:
            ts = copy.deepcopy(self.titles.pages[self.pageIndex][self.tsIndex])
            ts.y += util.getTextHeight(ts.size)
        else:
            ts = titles.TitleString(["new string"], 0.0, 100.0)

        self.titles.pages[self.pageIndex].append(ts)
        self.tsIndex = len(self.titles.pages[self.pageIndex]) - 1

        self.updateGui()

    def OnDeleteString(self, event):
        if (self.pageIndex == -1) or (self.tsIndex == -1):
            return

        del self.titles.pages[self.pageIndex][self.tsIndex]
        self.tsIndex = min(self.tsIndex, len(self.titles.pages[self.pageIndex]) - 1)

        self.updateGui()

    # update page/string listboxes and selection
    def updateGui(self):
        self.stringsLb.Clear()

        pgCnt = len(self.titles.pages)

        self.delPageBtn.Enable(pgCnt > 0)
        self.moveBtn.Enable(pgCnt > 1)
        self.nextBtn.Enable(pgCnt > 1)
        self.previewBtn.Enable(pgCnt > 0)

        if self.pageIndex != -1:
            page = self.titles.pages[self.pageIndex]

            self.pageLabel.SetLabel("Page: %d / %d" % (self.pageIndex + 1, pgCnt))
            self.addBtn.Enable(True)
            self.delBtn.Enable(len(page) > 0)

            for s in page:
                self.stringsLb.Append("--".join(s.items))

            if self.tsIndex != -1:
                self.stringsLb.SetSelection(self.tsIndex)
        else:
            self.pageLabel.SetLabel("No pages.")
            self.addBtn.Disable()
            self.delBtn.Disable()

        self.updateStringGui()

        self.previewCtrl.Refresh()

    # update selected string stuff
    def updateStringGui(self):
        if self.tsIndex == -1:
            for w in self.widList:
                w.Disable()

            self.textEntry.SetValue("")
            self.xEntry.SetValue("")
            self.yEntry.SetValue("")
            self.sizeEntry.SetValue(12)
            self.boldCb.SetValue(False)
            self.italicCb.SetValue(False)
            self.underlinedCb.SetValue(False)

            return

        self.block = True

        ts = self.titles.pages[self.pageIndex][self.tsIndex]

        for w in self.widList:
            w.Enable(True)

        if ts.isCentered:
            self.xEntry.Disable()

        self.textEntry.SetValue("\n".join(ts.items))

        self.xEntry.SetValue("%.2f" % ts.x)
        self.yEntry.SetValue("%.2f" % ts.y)

        util.reverseComboSelect(self.alignCombo, ts.getAlignment())

        util.reverseComboSelect(self.fontCombo, ts.font)
        self.sizeEntry.SetValue(ts.size)

        self.boldCb.SetValue(ts.isBold)
        self.italicCb.SetValue(ts.isItalic)
        self.underlinedCb.SetValue(ts.isUnderlined)

        self.block = False

        self.previewCtrl.Refresh()

    def OnMisc(self, event=None):
        if (self.tsIndex == -1) or self.block:
            return

        ts = self.titles.pages[self.pageIndex][self.tsIndex]

        ts.items = [
            util.toInputStr(s)
            for s in misc.fromGUI(self.textEntry.GetValue()).split("\n")
        ]

        self.stringsLb.SetString(self.tsIndex, "--".join(ts.items))

        ts.x = util.str2float(self.xEntry.GetValue(), 0.0)
        ts.y = util.str2float(self.yEntry.GetValue(), 0.0)

        ts.setAlignment(self.alignCombo.GetClientData(self.alignCombo.GetSelection()))
        self.xEntry.Enable(not ts.isCentered)

        ts.size = util.getSpinValue(self.sizeEntry)
        ts.font = self.fontCombo.GetClientData(self.fontCombo.GetSelection())

        ts.isBold = self.boldCb.GetValue()
        ts.isItalic = self.italicCb.GetValue()
        ts.isUnderlined = self.underlinedCb.GetValue()

        self.previewCtrl.Refresh()



# -*- coding: utf-8 -*-

import functools
import glob
import gzip
import os
import re
import tempfile
import time
from typing import AnyStr, Optional

import trelby.config as config
import trelby.misc as misc

if "TRELBY_TESTING" in os.environ:
    import unittest.mock as mock

    wx = mock.Mock()
else:
    import wx

# alignment values
ALIGN_LEFT = 0
ALIGN_CENTER = 1
ALIGN_RIGHT = 2
VALIGN_TOP = 1
VALIGN_CENTER = 2
VALIGN_BOTTOM = 3

# this has to be below the ALIGN stuff, otherwise things break due to
# circular dependencies
import trelby.fontinfo as fontinfo

# mappings from lowercase to uppercase letters for different charsets
_iso_8859_1_map = {
    97: 65,
    98: 66,
    99: 67,
    100: 68,
    101: 69,
    102: 70,
    103: 71,
    104: 72,
    105: 73,
    106: 74,
    107: 75,
    108: 76,
    109: 77,
    110: 78,
    111: 79,
    112: 80,
    113: 81,
    114: 82,
    115: 83,
    116: 84,
    117: 85,
    118: 86,
    119: 87,
    120: 88,
    121: 89,
    122: 90,
    224: 192,
    225: 193,
    226: 194,
    227: 195,
    228: 196,
    229: 197,
    230: 198,
    231: 199,
    232: 200,
    233: 201,
    234: 202,
    235: 203,
    236: 204,
    237: 205,
    238: 206,
    239: 207,
    240: 208,
    241: 209,
    242: 210,
    243: 211,
    244: 212,
    245: 213,
    246: 214,
    248: 216,
    249: 217,
    250: 218,
    251: 219,
    252: 220,
    253: 221,
    254: 222,
}

# current mappings, 256 chars long.
_to_upper = ""
_to_lower = ""

# translate table for converting strings to only contain valid input
# characters
_input_tbl = ""

# translate table that converts A-Z -> a-z, keeps a-z as they are, and
# converts everything else to z.
_normalize_tbl = ""

# identity table that maps each character to itself. used by deleteChars.
_identity_tbl = ""

# map some fancy unicode characters to their nearest ASCII/Latin-1
# equivalents so when people import text it's not mangled to uselessness
_fancy_unicode_map = {
    #    ord('‘') : "'",
    #    ord('’') : "'",
    #    ord('“') : '"',
    #    ord('”') : '"',
    #    ord('—') : "--",
    #    ord('–') : "-",
}

# permanent memory DC to get text extents etc
permDc = None


def init(doWX=True):
    global _to_upper, _to_lower, _input_tbl, _normalize_tbl, _identity_tbl, permDc

    # setup ISO-8859-1 case-conversion stuff
    tmpUpper = []
    tmpLower = []

    for i in range(256):
        tmpUpper.append(i)
        tmpLower.append(i)

    for k, v in _iso_8859_1_map.items():
        tmpUpper[k] = v
        tmpLower[v] = k

    for i in range(256):
        _to_upper += chr(tmpUpper[i])
        _to_lower += chr(tmpLower[i])

    # valid input string stuff
    for i in range(256):
        if isValidInputChar(i):
            _input_tbl += chr(i)
        else:
            _input_tbl += "|"

    for i in range(256):
        # "a" - "z"
        if (i >= 97) and (i <= 122):
            ch = chr(i)
        # "A" - "Z"
        elif (i >= 65) and (i <= 90):
            # + 32 ("A" - "a") lowercases it
            ch = chr(i + 32)
        else:
            ch = "z"

        _normalize_tbl += ch

    _identity_tbl = "".join([chr(i) for i in range(256)])

    if doWX:
        # dunno if the bitmap needs to be big enough to contain the text
        # we're measuring...
        permDc = wx.MemoryDC()
        permDc.SelectObject(wx.Bitmap(512, 32))


# like string.upper/lower/capitalize, but we do our own charset-handling
# that doesn't need locales etc
def upper(s):
    return s.upper()


def lower(s):
    return s.lower()


def capitalize(s):
    return upper(s[:1]) + s[1:]


# return 's', which must be a unicode string, converted to a ISO-8859-1
# 8-bit string. characters not representable in ISO-8859-1 are discarded.
def toLatin1(s):
    return s.encode("ISO-8859-1", "ignore")


# return 's', which must be a string of ISO-8859-1 characters, converted
# to UTF-8.
def toUTF8(s):
    return s


# return 's', which must be a string of UTF-8 characters, converted to
# ISO-8859-1, with characters not representable in ISO-8859-1 discarded
# and any invalid UTF-8 sequences ignored.
def fromUTF8(s):
    return s


# returns True if kc (key-code) is a valid character to add to the script.
def isValidInputChar(kc):
    # [0x80, 0x9F] = unspecified control characters in ISO-8859-1, added
    # characters like euro etc in windows-1252. 0x7F = backspace, 0xA0 =
    # non-breaking space, 0xAD = soft hyphen.
    return (
        (kc >= 32)
        and (kc <= 255)
        and not ((kc >= 0x7F) and (kc < 0xA0))
        and (kc != 0xAD)
    )


# return s with all non-valid input characters converted to valid input
# characters, except form feeds, which are just deleted.
def toInputStr(s):
    return str(s).replace("\f", "").replace("\t", "|")


# replace fancy unicode characters with their ASCII/Latin1 equivalents.
def removeFancyUnicode(s):
    return s.translate(_fancy_unicode_map)


# transform external input (unicode) into a form suitable for having in a
# script
def cleanInput(s):
    return s


# replace s[start:start + width] with toInputStr(new) and return s
def replace(s, new, start, width):
    return s[0:start] + toInputStr(new) + s[start + width :]


# delete all characters in 'chars' from s and return that.
def deleteChars(s: str, chars: str) -> str:
    for char in chars:
        s = s.replace(char, "")
    return s


# returns s with all possible different types of newlines converted to
# unix newlines, i.e. a single "\n"
def fixNL(s):
    return s.replace("\r\n", "\n").replace("\r", "\n")


# clamps the given value to a specific range. both limits are optional.
def clamp(val, minVal=None, maxVal=None):
    ret = val

    if minVal != None:
        ret = max(ret, minVal)

    if maxVal != None:
        ret = min(ret, maxVal)

    return ret


# like clamp, but gets/sets value directly from given object
def clampObj(obj, name, minVal=None, maxVal=None):
    setattr(obj, name, clamp(getattr(obj, name), minVal, maxVal))


# convert given string to float, clamping it to the given range
# (optional). never throws any exceptions, return defVal (possibly clamped
# as well) on any errors.
def str2float(s, defVal, minVal=None, maxVal=None):
    val = defVal

    try:
        val = float(s)
    except (ValueError, OverflowError):
        pass

    return clamp(val, minVal, maxVal)


# like str2float, but for ints.
def str2int(s, defVal, minVal=None, maxVal=None, radix=10):
    val = defVal

    try:
        val = int(s, radix)
    except ValueError:
        pass

    return clamp(val, minVal, maxVal)


# extract 'name' field from each item in 'seq', put it in a list, and
# return that list.
def listify(seq, name):
    l = []
    for it in seq:
        l.append(getattr(it, name))

    return l


# return percentage of 'val1' of 'val2' (both ints) as an int (50% -> 50
# etc.), or 0 if val2 is 0.
def pct(val1, val2):
    if val2 != 0:
        return (100 * val1) // val2
    else:
        return 0


# return percentage of 'val1' of 'val2' (both ints/floats) as a float (50%
# -> 50.0 etc.), or 0.0 if val2 is 0.0
def pctf(val1, val2):
    if val2 != 0.0:
        return (100.0 * val1) / val2
    else:
        return 0.0


# return float(val1) / val2, or 0.0 if val2 is 0.0
def safeDiv(val1, val2):
    if val2 != 0.0:
        return float(val1) / val2
    else:
        return 0.0


# return float(val1) / val2, or 0.0 if val2 is 0
def safeDivInt(val1, val2):
    if val2 != 0:
        return float(val1) / val2
    else:
        return 0.0


# for each character in 'flags', starting at beginning, checks if that
# character is found in 's'. if so, appends True to a tuple, False
# otherwise. returns that tuple, whose length is of course is len(flags).
def flags2bools(s, flags):
    b = ()

    for f in flags:
        if s.find(f) != -1:
            b += (True,)
        else:
            b += (False,)

    return b


# reverse of flags2bools. is given a number of objects, if each object
# evaluates to true, chars[i] is appended to the return string. len(chars)
# == len(bools) must be true.
def bools2flags(chars, *bools):
    s = ""

    if len(chars) != len(bools):
        raise TypeError("bools2flags: chars and bools are not equal length")

    for i in range(len(chars)):
        if bools[i]:
            s += chars[i]

    return s


# return items, which is a list of ISO-8859-1 strings, as a single string
# with \n between each string. any \ characters in the individual strings
# are escaped as \\.
def escapeStrings(items):
    return "\\n".join([s.replace("\\", "\\\\") for s in items])


# opposite of escapeStrings. takes in a string, returns a list of strings.
def unescapeStrings(s):
    if not s:
        return []

    items = []

    tmp = ""
    i = 0
    while i < (len(s) - 1):
        ch = s[i]

        if ch != "\\":
            tmp += ch
            i += 1
        else:
            ch = s[i + 1]

            if ch == "n":
                items.append(tmp)
                tmp = ""
            else:
                tmp += ch

            i += 2

    if i < len(s):
        tmp += s[i]
        items.append(tmp)

    return items


# return s encoded so that all characters outside the range [32,126] (and
# "\\") are escaped.
def encodeStr(s):
    ret = ""

    for ch in s:
        c = ord(ch)

        # ord("\\") == 92 == 0x5C
        if c == 92:
            ret += "\\5C"
        elif (c >= 32) and (c <= 126):
            ret += ch
        else:
            ret += "\\%02X" % c

    return ret


# reverse of encodeStr. if string contains invalid escapes, they're
# silently and arbitrarily replaced by something.
def decodeStr(s):
    return re.sub(r"\\..", _decodeRepl, s)


# converts "\A4" style matches to their character values.
def _decodeRepl(mo):
    val = str2int(mo.group(0)[1:], 256, 0, 256, 16)

    if val != 256:
        return chr(val)
    else:
        return ""


# return string s escaped for use in RTF.
def escapeRTF(s):
    return s.replace("\\", "\\\\").replace("{", r"\{").replace("}", r"\}")


# convert mm to twips (1/1440 inch = 1/20 point).
def mm2twips(mm):
    # 56.69291 = 1440 / 25.4
    return mm * 56.69291


# TODO: move all GUI stuff to gutil


# return True if given font is a fixed-width one.
def isFixedWidth(font):
    return getTextExtent(font, "iiiii")[0] == getTextExtent(font, "OOOOO")[0]


# get extent of 's' as (w, h)
def getTextExtent(font, s):
    permDc.SetFont(font)

    # if we simply return permDc.GetTextExtent(s) from here, on some
    # versions of Windows we will incorrectly reject as non-fixed width
    # fonts (through isFixedWidth) some fonts that actually are fixed
    # width. it's especially bad because one of them is our default font,
    # "Courier New".
    #
    # these are the widths we get for the strings below for Courier New, italic:
    #
    # iiiii 40
    # iiiiiiiiii 80
    # OOOOO 41
    # OOOOOOOOOO 81
    #
    # we can see i and O are both 8 pixels wide, so the font is
    # fixed-width, but for whatever reason, on the O variants there is one
    # additional pixel returned in the width, no matter what the length of
    # the string is.
    #
    # to get around this, we actually call GetTextExtent twice, once with
    # the actual string we want to measure, and once with the string
    # duplicated, and take the difference between those two as the actual
    # width. this handily negates the one-extra-pixel returned and gives
    # us an accurate way of checking if a font is fixed width or not.
    #
    # it's a bit slower but this is not called from anywhere that's
    # performance critical.

    w1, h = permDc.GetTextExtent(s)
    w2 = permDc.GetTextExtent(s + s)[0]

    return (w2 - w1, h)


# get height of font in pixels
def getFontHeight(font):
    permDc.SetFont(font)
    return permDc.GetTextExtent("_\xC5")[1]


# return how many mm tall given font size is.
def getTextHeight(size):
    return (size / 72.0) * 25.4


# return how many mm wide given text is at given style with given size.
def getTextWidth(text, style, size):
    return (fontinfo.getMetrics(style).getTextWidth(text, size) / 72.0) * 25.4


# create a font that's height is at most 'height' pixels. other parameters
# are the same as in wx.Font's constructor.
def createPixelFont(height, family, style, weight):
    fs = 6

    selected = fs
    closest = 1000
    over = 0

    # FIXME: what's this "keep trying even once we go over the max height"
    # stuff? get rid of it.
    while 1:
        fn = wx.Font(fs, family, style, weight, encoding=wx.FONTENCODING_ISO8859_1)
        h = getFontHeight(fn)
        diff = height - h

        if diff >= 0:
            if diff < closest:
                closest = diff
                selected = fs
        else:
            over += 1

        if (over >= 3) or (fs > 144):
            break

        fs += 2

    return wx.Font(selected, family, style, weight, encoding=wx.FONTENCODING_ISO8859_1)


def reverseComboSelect(combo, clientData):
    for i in range(combo.GetCount()):
        if combo.GetClientData(i) == clientData:
            if combo.GetSelection() != i:
                combo.SetSelection(i)

            return True

    return False


# set widget's client size. if w or h is -1, that dimension is not changed.
def setWH(ctrl, w=-1, h=-1):
    size = ctrl.GetClientSize()

    if w != -1:
        size.width = w

    if h != -1:
        size.height = h

    ctrl.SetMinSize(wx.Size(size.width, size.height))
    ctrl.SetClientSize(size.width, size.height)


# wxMSW doesn't respect the control's min/max values at all, so we have to
# implement this ourselves
def getSpinValue(spinCtrl):
    tmp = clamp(spinCtrl.GetValue(), spinCtrl.GetMin(), spinCtrl.GetMax())
    spinCtrl.SetValue(tmp)

    return tmp


# return True if c is not a word character, i.e. is either empty, not an
# alphanumeric character or a "'", or is more than one character.
def isWordBoundary(c):
    if len(c) != 1:
        return True

    if c == "'":
        return False

    return not isAlnum(c)


# return True if c is an alphanumeric character
def isAlnum(c):
    return str(c).isalnum()


# make sure s (unicode) ends in suffix (case-insensitively) and return
# that. suffix must already be lower-case.
def ensureEndsIn(s, suffix):
    if s.lower().endswith(suffix):
        return s
    else:
        return s + suffix


# return string 's' split into words (as a list), using isWordBoundary.
def splitToWords(s):
    tmp = ""

    for c in s:
        if isWordBoundary(c):
            tmp += " "
        else:
            tmp += c

    return tmp.split()


# return two-character prefix of s, using characters a-z only. len(s) must
# be at least 2.
def getWordPrefix(s):
    return s[:2].translate(_normalize_tbl)


# return count of how many 'ch' characters 's' begins with.
def countInitial(s, ch):
    cnt = 0

    for i in range(len(s)):
        if s[i] != ch:
            break

        cnt += 1

    return cnt


# searches string 's' for each item of list 'seq', returning True if any
# of them were found.
def multiFind(s, seq):
    for it in seq:
        if s.find(it) != -1:
            return True

    return False


def cmpfunc(a, b):
    return (a > b) - (a < b)


# put everything from dictionary d into a list as (key, value) tuples,
# then sort the list and return that. by default sorts by "desc(value)
# asc(key)", but a custom sort function can be given
def sortDict(d, sortFunc=None):
    def tmpSortFunc(o1, o2):
        ret = cmpfunc(o2[1], o1[1])

        if ret != 0:
            return ret
        else:
            return cmpfunc(o1[0], o2[0])

    if sortFunc == None:
        sortFunc = tmpSortFunc

    tmp = []
    for k, v in d.items():
        tmp.append((k, v))

    tmp = sorted(tmp, key=functools.cmp_to_key(sortFunc))

    return tmp


# an efficient FIFO container of fixed size. can't contain None objects.
class FIFO:
    def __init__(self, size):
        self.arr = [None] * size

        # index of next slot to fill
        self.next = 0

    # add item
    def add(self, obj):
        self.arr[self.next] = obj
        self.next += 1

        if self.next >= len(self.arr):
            self.next = 0

    # get contents as a list, in LIFO order.
    def get(self):
        tmp = []

        j = self.next - 1

        for i in range(len(self.arr)):
            if j < 0:
                j = len(self.arr) - 1

            obj = self.arr[j]

            if obj != None:
                tmp.append(obj)

            j -= 1

        return tmp


# DrawLine-wrapper that makes it easier when the end-point is just
# offsetted from the starting point
def drawLine(dc, x, y, xd, yd):
    dc.DrawLine(x, y, x + xd, y + yd)


# draws text aligned somehow. returns a (w, h) tuple of the text extent.
def drawText(dc, text, x, y, align=ALIGN_LEFT, valign=VALIGN_TOP):
    w, h = dc.GetTextExtent(text)

    if align == ALIGN_CENTER:
        x -= w // 2
    elif align == ALIGN_RIGHT:
        x -= w

    if valign == VALIGN_CENTER:
        y -= h // 2
    elif valign == VALIGN_BOTTOM:
        y -= h

    dc.DrawText(text, x, y)

    return (w, h)


# create pad sizer for given window whose controls are in topSizer, with
# 'pad' pixels of padding on each side, resize window to correct size, and
# optionally center it.
def finishWindow(window, topSizer, pad=10, center=False):
    padSizer = wx.BoxSizer(wx.VERTICAL)
    padSizer.Add(topSizer, 1, wx.EXPAND | wx.ALL, pad)
    window.SetSizerAndFit(padSizer)
    window.Layout()

    if center:
        window.Center()


# wx.Colour replacement that can safely be copy.deepcopy'd
class MyColor:
    def __init__(self, r, g, b):
        self.r = r
        self.g = g
        self.b = b

    def toWx(self):
        return wx.Colour(self.r, self.g, self.b)

    @staticmethod
    def fromWx(c):
        o = MyColor(0, 0, 0)

        o.r = c.Red()
        o.g = c.Green()
        o.b = c.Blue()

        return o


# fake key event, supports same operations as the real one
class MyKeyEvent:
    def __init__(self, kc=0):
        # keycode
        self.kc = kc

        self.controlDown = False
        self.altDown = False
        self.shiftDown = False

    def GetKeyCode(self):
        return self.kc

    def ControlDown(self):
        return self.controlDown

    def AltDown(self):
        return self.altDown

    def ShiftDown(self):
        return self.shiftDown

    def Skip(self):
        pass


# one key press
class Key:
    keyMap = {
        1: "A",
        2: "B",
        3: "C",
        4: "D",
        5: "E",
        6: "F",
        7: "G",
        # CTRL+Enter = 10 in Windows
        10: "Enter (Windows)",
        11: "K",
        12: "L",
        14: "N",
        15: "O",
        16: "P",
        17: "Q",
        18: "R",
        19: "S",
        20: "T",
        21: "U",
        22: "V",
        23: "W",
        24: "X",
        25: "Y",
        26: "Z",
        wx.WXK_BACK: "Backspace",
        wx.WXK_TAB: "Tab",
        wx.WXK_RETURN: "Enter",
        wx.WXK_ESCAPE: "Escape",
        wx.WXK_DELETE: "Delete",
        wx.WXK_END: "End",
        wx.WXK_HOME: "Home",
        wx.WXK_LEFT: "Left",
        wx.WXK_UP: "Up",
        wx.WXK_RIGHT: "Right",
        wx.WXK_DOWN: "Down",
        wx.WXK_PAGEUP: "Page up",
        wx.WXK_PAGEDOWN: "Page down",
        wx.WXK_INSERT: "Insert",
        wx.WXK_F1: "F1",
        wx.WXK_F2: "F2",
        wx.WXK_F3: "F3",
        wx.WXK_F4: "F4",
        wx.WXK_F5: "F5",
        wx.WXK_F6: "F6",
        wx.WXK_F7: "F7",
        wx.WXK_F8: "F8",
        wx.WXK_F9: "F9",
        wx.WXK_F10: "F10",
        wx.WXK_F11: "F11",
        wx.WXK_F12: "F12",
        wx.WXK_F13: "F13",
        wx.WXK_F14: "F14",
        wx.WXK_F15: "F15",
        wx.WXK_F16: "F16",
        wx.WXK_F17: "F17",
        wx.WXK_F18: "F18",
        wx.WXK_F19: "F19",
        wx.WXK_F20: "F20",
        wx.WXK_F21: "F21",
        wx.WXK_F22: "F22",
        wx.WXK_F23: "F23",
        wx.WXK_F24: "F24",
    }

    def __init__(self, kc, ctrl=False, alt=False, shift=False):

        # we don't want to handle ALT+a/ALT+A etc separately, so uppercase
        # input char combinations
        if (kc < 256) and (ctrl or alt):
            kc = ord(upper(chr(kc)))

        # even though the wxWidgets documentation clearly states that
        # CTRL+[A-Z] should be returned as keycodes 1-26, wxGTK2 2.6 does
        # not do this (wxGTK1 and wxMSG do follow the documentation).
        #
        # so, we normalize to the wxWidgets official form here if necessary.

        # "A" - "Z"
        if ctrl and (kc >= 65) and (kc <= 90):
            kc -= 64

        # ASCII/Latin-1 keycode (0-255) or one of the wx.WXK_ constants (>255)
        self.kc = kc

        self.ctrl = ctrl
        self.alt = alt
        self.shift = shift

    # returns True if key is a valid input character
    def isValidInputChar(self):
        return not self.ctrl and not self.alt and isValidInputChar(self.kc)

    # toInt/fromInt serialize/deserialize to/from a 35-bit integer, laid
    # out like this:
    # bits 0-31:  keycode
    #        32:  Control
    #        33:  Alt
    #        34:  Shift

    def toInt(self):
        return (
            (self.kc & 0xFFFFFFFF)
            | (self.ctrl << 32)
            | (self.alt << 33)
            | (self.shift << 34)
        )

    @staticmethod
    def fromInt(val):
        return Key(val & 0xFFFFFFFF, (val >> 32) & 1, (val >> 33) & 1, (val >> 34) & 1)

    # construct from wx.KeyEvent
    @staticmethod
    def fromKE(ev):
        return Key(ev.GetKeyCode(), ev.ControlDown(), ev.AltDown(), ev.ShiftDown())

    def toStr(self):
        s = ""

        if self.ctrl:
            s += "CTRL+"

        if self.alt:
            s += "ALT+"

        if self.shift:
            s += "SHIFT+"

        if isValidInputChar(self.kc):
            if self.kc == wx.WXK_SPACE:
                s += "Space"
            else:
                s += chr(self.kc)
        else:
            kname = self.__class__.keyMap.get(self.kc)

            if kname:
                s += kname
            else:
                s += "UNKNOWN(%d)" % self.kc

        return s


# a string-like object that features reasonably fast repeated appends even
# for large strings, since it keeps each appended string as an item in a
# list.
class String:
    def __init__(self, s=None):

        # byte count of data appended
        self.pos = 0

        # list of strings
        self.data = []

        if s:
            self += s

    def __len__(self):
        return self.pos

    def __str__(self):
        return "".join(self.data)

    def __iadd__(self, s):
        s2 = str(s)

        self.data.append(s2)
        self.pos += len(s2)

        return self


# load at most maxSize (all if -1) bytes from 'filename', returning the
# data as a string or None on errors. pops up message boxes with 'frame'
# as parent on errors.
def loadFile(
    filename: str, frame: Optional[wx.Frame], maxSize: int = -1, binary: bool = False
) -> Optional[AnyStr]:
    try:
        if binary:
            f = open(misc.toPath(filename), "rb")
        else:
            f = open(misc.toPath(filename), "r", encoding="UTF-8")

        try:
            ret = f.read(maxSize)
        except MemoryError as error:
            print(f"File too large: {error}")
        finally:
            f.close()

    except IOError as xxx_todo_changeme:
        (errno, strerror) = xxx_todo_changeme.args
        wx.MessageBox(
            "Error loading file '%s': %s" % (filename, strerror), "Error", wx.OK, frame
        )
        ret = None

    return ret


def loadLatin1EncodedFile(
    filename: str, frame: Optional[wx.Frame], maxSize: int = -1
) -> Optional[AnyStr]:
    return loadFile(filename, frame, maxSize, "ISO-8859-1")


# like loadFile, but if file doesn't exist, tries to load a .gz compressed
# version of it.
def loadMaybeCompressedFile(filename, frame):
    doGz = False

    if not fileExists(filename):
        filename += ".gz"
        doGz = True

    if not doGz:
        s = loadFile(filename, frame)
        return s

    try:
        f = gzip.open(filename, "r")
        return f.read().decode("utf-8")
    except:
        wx.MessageBox(
            "Error loading file '%s': Decompression failed" % filename,
            "Error",
            wx.OK,
            frame,
        )
        return None


# write 'data' to 'filename', popping up a messagebox using 'frame' as
# parent on errors. returns True on success.
def writeToFile(filename: str, data: AnyStr, frame: wx.TopLevelWindow) -> bool:
    try:
        f = open(misc.toPath(filename), "wb")

        try:
            if isinstance(data, str):
                f.write(data.encode("UTF-8"))
            else:
                f.write(data)
        finally:
            f.close()

        return True

    except IOError as xxx_todo_changeme1:
        (errno, strerror) = xxx_todo_changeme1.args
        wx.MessageBox(
            "Error writing file '%s': %s" % (filename, strerror), "Error", wx.OK, frame
        )

        return False


def removeTempFiles(prefix):
    files = glob.glob(tempfile.gettempdir() + "/%s*" % prefix)

    for fn in files:
        try:
            os.remove(fn)
        except OSError:
            continue


# return True if given file exists.
def fileExists(filename):
    try:
        os.stat(misc.toPath(filename))
    except OSError:
        return False

    return True


# look for file 'filename' in all the directories listed in 'dirs', which
# is a list of absolute directory paths. if found, return the absolute
# filename, otherwise None.
def findFile(filename, dirs):
    for d in dirs:
        if d[-1] != "/":
            d += "/"

        path = d + filename

        if fileExists(path):
            return path

    return None


# look for file 'filename' in all the directories listed in $PATH. if
# found, return the absolute filename, otherwise None.
def findFileInPath(filename):
    dirs = os.getenv("PATH")
    if not dirs:
        return None

    # I have no idea how one should try to cope if PATH contains entries
    # with non-UTF8 characters, so just ignore any errors
    dirs = str(dirs).split(":")

    # only accept absolute paths. this strips out things like "~/bin/"
    # etc.
    dirs = [d for d in dirs if d and d[0] == "/"]

    return findFile(filename, dirs)


# simple timer class for use during development only
class TimerDev:

    # how many TimerDev instances are currently in existence
    nestingLevel = 0

    def __init__(self, msg=""):
        self.msg = msg
        self.__class__.nestingLevel += 1
        self.t = time.time()

    def __del__(self):
        self.t = time.time() - self.t
        self.__class__.nestingLevel -= 1
        print(
            "%s%s took %.5f seconds"
            % (" " * self.__class__.nestingLevel, self.msg, self.t)
        )


# Get the Windows default PDF viewer path from registry and return that,
# or None on errors.
def _getWindowsPDFViewer():
    try:
        import winreg

        # HKCR/.pdf: gives the class of the PDF program.
        # Example : AcroRead.Document or FoxitReader.Document

        key = winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, ".pdf")
        pdfClass = winreg.QueryValue(key, "")

        # HKCR/<class>/shell/open/command: the path to the PDF viewer program
        # Example: "C:\Program Files\Acrobat 8.0\acroread.exe" "%1"

        key2 = winreg.OpenKey(
            winreg.HKEY_CLASSES_ROOT, pdfClass + r"\shell\open\command"
        )

        # Almost every PDF program out there accepts passing the PDF path
        # as the argument, so we don't parse the arguments from the
        # registry, just get the program path.

        path = winreg.QueryValue(key2, "").split('"')[1]

        if fileExists(path):
            return path
    except:
        pass

    return None


# Get a tuple of the default PDF viewer path and args for the current platform
# and return that, or a tuple of two Nones on errors.
def getPDFViewer():
    # list of programs to look for. each item is of the form (name,
    # args). if name is an absolute path only that exact location is
    # looked at, otherwise PATH is searched for the program (on
    # Windows, all paths are interpreted as absolute). args is the
    # list of arguments for the program.
    progs = []

    if misc.isMac:
        # Need to check for Mac before Unix as macOS is a Unix
        progs = [("/System/Applications/Preview.app", "")]
    elif misc.isUnix:
        progs = [("/usr/bin/xdg-open/", "")]
    elif misc.isWindows:
        # get value via registry if possible, or fallback to old method.
        viewer = _getWindowsPDFViewer()
        if viewer:
            return (viewer, "")

        progs = [
            (r"C:\Program Files\Adobe\Reader 11.0\Reader\AcroRd32.exe", ""),
            (r"C:\Program Files\Adobe\Reader 10.0\Reader\AcroRd32.exe", ""),
            (r"C:\Program Files\Adobe\Reader 9.0\Reader\AcroRd32.exe", ""),
            (r"C:\Program Files\Adobe\Reader 8.0\Reader\AcroRd32.exe", ""),
            (r"C:\Program Files\Adobe\Acrobat 7.0\Reader\AcroRd32.exe", ""),
            (r"C:\Program Files\Adobe\Acrobat 6.0\Reader\AcroRd32.exe", ""),
            (r"C:\Program Files\Adobe\Acrobat 5.0\Reader\AcroRd32.exe", ""),
            (r"C:\Program Files\Adobe\Acrobat 4.0\Reader\AcroRd32.exe", ""),
            (r"C:\Program Files\Foxit Software\Foxit Reader\Foxit Reader.exe", ""),
        ]
    else:
        # Unknown platform, don't know where to look
        pass
    for name, args in progs:
        if misc.isWindows or (name[0] == "/"):
            if fileExists(name):
                return (name, args)
        else:
            name = findFileInPath(name)
            if name:
                return (name, args)
    return (None, None)


# get a windows environment variable in its native unicode format, or None
# if not found
def getWindowsUnicodeEnvVar(name):
    import ctypes

    n = ctypes.windll.kernel32.GetEnvironmentVariableW(name, None, 0)

    if n == 0:
        return None

    buf = ctypes.create_unicode_buffer("\0" * n)
    ctypes.windll.kernel32.GetEnvironmentVariableW(name, buf, n)

    return buf.value


# show PDF file.
def showPDF(
    filename: str, cfgGl: "config.ConfigGlobal", frame: wx.TopLevelWindow
) -> None:
    def complain():
        wx.MessageBox(
            "PDF viewer application not found.\n\n"
            "You can change your PDF viewer\n"
            "settings at File/Settings/Change/Misc.",
            "Error",
            wx.OK,
            frame,
        )

    pdfProgram = cfgGl.pdfViewerPath
    pdfArgs = cfgGl.pdfViewerArgs

    # If configured pdf viewer does not exist, try finding one
    # automatically
    if not fileExists(pdfProgram):
        found = False

        regPDF, regArgs = getPDFViewer()

        if regPDF:
            wx.MessageBox(
                "Currently set PDF viewer (%s) was not found.\n"
                "Change this in File/Settings/Change/Misc.\n\n"
                "Using the default PDF viewer instead:\n"
                "%s" % (pdfProgram, regPDF),
                "Warning",
                wx.OK,
                frame,
            )

            pdfProgram = regPDF
            pdfArgs = regArgs

            found = True

        if not found:
            complain()

            return

    if misc.isMac:
        # on Mac, we should use 'open', so wrap the whole call in that
        args = ["open", "-a", pdfProgram, filename]
        if pdfArgs:
            args = args + ["--args"] + pdfArgs.split()
        pdfProgram = "/usr/bin/open"
    else:
        # on Windows, Acrobat complains about "invalid path" if we
        # give the full path of the program as first arg, so give a
        # dummy arg.
        args = ["pdf"] + pdfArgs.split() + [filename]

    # there's a race condition in checking if the path exists, above, and
    # using it, below. if the file disappears between those two we get an
    # OSError exception from spawnv, so we need to catch it and handle it.

    # TODO: spawnv does not support Unicode paths as of this moment
    # (Python 2.4). for now, convert it to UTF-8 and hope for the best.
    try:
        os.spawnv(os.P_NOWAIT, pdfProgram.encode("UTF-8"), args)
    except OSError:
        complain()



# see fileformat.txt for more detailed information about the various
# defines found here.

import copy
import os

import trelby.misc as misc
import trelby.mypickle as mypickle
import trelby.pml as pml
import trelby.screenplay as screenplay
import trelby.util as util
from trelby.error import ConfigError

if "TRELBY_TESTING" in os.environ:
    import unittest.mock as mock

    wx = mock.Mock()
else:
    import wx

# mapping from character to linebreak
_char2lb = {
    ">": screenplay.LB_SPACE,
    "+": screenplay.LB_SPACE2,
    "&": screenplay.LB_NONE,
    "|": screenplay.LB_FORCED,
    ".": screenplay.LB_LAST,
}

# reverse to above
_lb2char = {}

# what string each linebreak type should be mapped to.
_lb2str = {
    screenplay.LB_SPACE: " ",
    screenplay.LB_SPACE2: "  ",
    screenplay.LB_NONE: "",
    screenplay.LB_FORCED: "\n",
    screenplay.LB_LAST: "\n",
}

# contains a TypeInfo for each element type
_ti = []

# mapping from character to TypeInfo
_char2ti = {}

# mapping from line type to TypeInfo
_lt2ti = {}

# mapping from element name to TypeInfo
_name2ti = {}

# page break indicators. do not change these values as they're saved to
# the config file.
PBI_NONE = 0
PBI_REAL = 1
PBI_REAL_AND_UNADJ = 2

# for range checking above value
PBI_FIRST, PBI_LAST = PBI_NONE, PBI_REAL_AND_UNADJ

# constants for identifying PDFFontInfos
PDF_FONT_NORMAL = "Normal"
PDF_FONT_BOLD = "Bold"
PDF_FONT_ITALIC = "Italic"
PDF_FONT_BOLD_ITALIC = "Bold-Italic"

# scrolling  directions
SCROLL_UP = 0
SCROLL_DOWN = 1
SCROLL_CENTER = 2

# construct reverse lookup tables

for k, v in list(_char2lb.items()):
    _lb2char[v] = k

del k, v


# non-changing information about an element type
class TypeInfo:
    def __init__(self, lt, char, name):

        # line type, e.g. screenplay.ACTION
        self.lt = lt

        # character used in saved scripts, e.g. "."
        self.char = char

        # textual name, e.g. "Action"
        self.name = name


# text type
class TextType:
    cvars = None

    def __init__(self):
        if not self.__class__.cvars:
            v = self.__class__.cvars = mypickle.Vars()

            v.addBool("isCaps", False, "AllCaps")
            v.addBool("isBold", False, "Bold")
            v.addBool("isItalic", False, "Italic")
            v.addBool("isUnderlined", False, "Underlined")

        self.__class__.cvars.setDefaults(self)

    def save(self, prefix):
        return self.cvars.save(prefix, self)

    def load(self, vals, prefix):
        self.cvars.load(vals, prefix, self)


# script-specific information about an element type
class Type:
    cvars = None

    def __init__(self, lt):

        # line type
        self.lt = lt

        # pointer to TypeInfo
        self.ti = lt2ti(lt)

        # text types, one for screen and one for export
        self.screen = TextType()
        self.export = TextType()

        if not self.__class__.cvars:
            v = self.__class__.cvars = mypickle.Vars()

            # these two are how much empty space to insert a) before the
            # element b) between the element's lines, in units of line /
            # 10.
            v.addInt("beforeSpacing", 0, "BeforeSpacing", 0, 50)
            v.addInt("intraSpacing", 0, "IntraSpacing", 0, 20)

            v.addInt("indent", 0, "Indent", 0, 80)
            v.addInt("width", 5, "Width", 5, 80)

            v.makeDicts()

        self.__class__.cvars.setDefaults(self)

    def save(self, prefix):
        prefix += "%s/" % self.ti.name

        s = self.cvars.save(prefix, self)
        s += self.screen.save(prefix + "Screen/")
        s += self.export.save(prefix + "Export/")

        return s

    def load(self, vals, prefix):
        prefix += "%s/" % self.ti.name

        self.cvars.load(vals, prefix, self)
        self.screen.load(vals, prefix + "Screen/")
        self.export.load(vals, prefix + "Export/")


# global information about an element type
class TypeGlobal:
    cvars = None

    def __init__(self, lt):

        # line type
        self.lt = lt

        # pointer to TypeInfo
        self.ti = lt2ti(lt)

        if not self.__class__.cvars:
            v = self.__class__.cvars = mypickle.Vars()

            # what type of element to insert when user presses enter or tab.
            v.addElemName("newTypeEnter", screenplay.ACTION, "NewTypeEnter")
            v.addElemName("newTypeTab", screenplay.ACTION, "NewTypeTab")

            # what element to switch to when user hits tab / shift-tab.
            v.addElemName("nextTypeTab", screenplay.ACTION, "NextTypeTab")
            v.addElemName("prevTypeTab", screenplay.ACTION, "PrevTypeTab")

            v.makeDicts()

        self.__class__.cvars.setDefaults(self)

    def save(self, prefix):
        prefix += "%s/" % self.ti.name

        return self.cvars.save(prefix, self)

    def load(self, vals, prefix):
        prefix += "%s/" % self.ti.name

        self.cvars.load(vals, prefix, self)


# command (an action in the main program)
class Command:
    cvars = None

    def __init__(
        self,
        name,
        desc,
        defKeys=[],
        isMovement=False,
        isFixed=False,
        isMenu=False,
        scrollDirection=SCROLL_CENTER,
    ):

        # name, e.g. "MoveLeft"
        self.name = name

        # textual description
        self.desc = desc

        # default keys (list of serialized util.Key objects (ints))
        self.defKeys = defKeys

        # is this a movement command
        self.isMovement = isMovement

        # some commands & their keys (Tab, Enter, Quit, etc) are fixed and
        # can't be changed
        self.isFixed = isFixed

        # is this a menu item
        self.isMenu = isMenu

        # which way the command wants to scroll the page
        self.scrollDirection = scrollDirection

        if not self.__class__.cvars:
            v = self.__class__.cvars = mypickle.Vars()

            v.addList(
                "keys", [], "Keys", mypickle.IntVar("", 0, "", 0, 9223372036854775808)
            )

            v.makeDicts()

        # this is not actually needed but let's keep it for consistency
        self.__class__.cvars.setDefaults(self)

        self.keys = copy.deepcopy(self.defKeys)

    def save(self, prefix):
        if self.isFixed:
            return ""

        prefix += "%s/" % self.name

        if len(self.keys) > 0:
            return self.cvars.save(prefix, self)
        else:
            self.keys.append(0)
            s = self.cvars.save(prefix, self)
            self.keys = []

            return s

    def load(self, vals, prefix):
        if self.isFixed:
            return

        prefix += "%s/" % self.name

        tmp = copy.deepcopy(self.keys)
        self.cvars.load(vals, prefix, self)

        if len(self.keys) == 0:
            # we have a new command in the program not found in the old
            # config file
            self.keys = tmp
        elif self.keys[0] == 0:
            self.keys = []

        # weed out invalid bindings
        tmp2 = self.keys
        self.keys = []

        for k in tmp2:
            k2 = util.Key.fromInt(k)
            if not k2.isValidInputChar():
                self.keys.append(k)


# information about one screen font
class FontInfo:
    def __init__(self):
        self.font = None

        # font width and height
        self.fx = 1
        self.fy = 1


# information about one PDF font
class PDFFontInfo:
    cvars = None

    # list of characters not allowed in pdfNames
    invalidChars = None

    def __init__(self, name, style):
        # our name for the font (one of the PDF_FONT_* constants)
        self.name = name

        # 2 lowest bits of pml.TextOp.flags
        self.style = style

        if not self.__class__.cvars:
            v = self.__class__.cvars = mypickle.Vars()

            # name to use in generated PDF file (CourierNew, MyFontBold,
            # etc.). if empty, use the default PDF Courier font.
            v.addStrLatin1("pdfName", "", "Name")

            # filename for the font to embed, or empty meaning don't
            # embed.
            v.addStrUnicode("filename", "", "Filename")

            v.makeDicts()

            tmp = ""

            for i in range(256):
                # the OpenType font specification 1.4, of all places,
                # contains the most detailed discussion of characters
                # allowed in Postscript font names, in the section on
                # 'name' tables, describing name ID 6 (=Postscript name).
                if (
                    (i <= 32)
                    or (i >= 127)
                    or chr(i) in ("[", "]", "(", ")", "{", "}", "<", ">", "/", "%")
                ):
                    tmp += chr(i)

            self.__class__.invalidChars = tmp

        self.__class__.cvars.setDefaults(self)

    def save(self, prefix):
        prefix += "%s/" % self.name

        return self.cvars.save(prefix, self)

    def load(self, vals, prefix):
        prefix += "%s/" % self.name

        self.cvars.load(vals, prefix, self)

    # fix up invalid values.
    def refresh(self):
        self.pdfName = util.deleteChars(self.pdfName, self.invalidChars)

        # to avoid confused users not understanding why their embedded
        # font isn't working, put in an arbitrary font name if needed
        if self.filename and not self.pdfName:
            self.pdfName = "SampleFontName"


# per-script config, each script has its own one of these.
class Config:
    cvars = None

    def __init__(self):

        if not self.__class__.cvars:
            self.setupVars()

        self.__class__.cvars.setDefaults(self)

        # type configs, key = line type, value = Type
        self.types = {}

        # element types
        t = Type(screenplay.SCENE)
        t.beforeSpacing = 10
        t.indent = 0
        t.width = 60
        t.screen.isCaps = True
        t.screen.isBold = True
        t.export.isCaps = True
        self.types[t.lt] = t

        t = Type(screenplay.ACTION)
        t.beforeSpacing = 10
        t.indent = 0
        t.width = 60
        self.types[t.lt] = t

        t = Type(screenplay.CHARACTER)
        t.beforeSpacing = 10
        t.indent = 22
        t.width = 38
        t.screen.isCaps = True
        t.export.isCaps = True
        self.types[t.lt] = t

        t = Type(screenplay.DIALOGUE)
        t.indent = 10
        t.width = 35
        self.types[t.lt] = t

        t = Type(screenplay.PAREN)
        t.indent = 16
        t.width = 25
        self.types[t.lt] = t

        t = Type(screenplay.TRANSITION)
        t.beforeSpacing = 10
        t.indent = 45
        t.width = 20
        t.screen.isCaps = True
        t.export.isCaps = True
        self.types[t.lt] = t

        t = Type(screenplay.SHOT)
        t.beforeSpacing = 10
        t.indent = 0
        t.width = 60
        t.screen.isCaps = True
        t.export.isCaps = True
        self.types[t.lt] = t

        t = Type(screenplay.ACTBREAK)
        t.beforeSpacing = 10
        t.indent = 25
        t.width = 10
        t.screen.isCaps = True
        t.screen.isBold = True
        t.screen.isUnderlined = True
        t.export.isCaps = True
        t.export.isUnderlined = True
        self.types[t.lt] = t

        t = Type(screenplay.NOTE)
        t.beforeSpacing = 10
        t.indent = 5
        t.width = 55
        t.screen.isItalic = True
        t.export.isItalic = True
        self.types[t.lt] = t

        # pdf font configs, key = PDF_FONT_*, value = PdfFontInfo
        self.pdfFonts = {}

        for name, style in (
            (PDF_FONT_NORMAL, pml.COURIER),
            (PDF_FONT_BOLD, pml.COURIER | pml.BOLD),
            (PDF_FONT_ITALIC, pml.COURIER | pml.ITALIC),
            (PDF_FONT_BOLD_ITALIC, pml.COURIER | pml.BOLD | pml.ITALIC),
        ):
            self.pdfFonts[name] = PDFFontInfo(name, style)

        self.recalc()

    def setupVars(self):
        v = self.__class__.cvars = mypickle.Vars()

        # font size used for PDF generation, in points
        v.addInt("fontSize", 12, "FontSize", 4, 72)

        # margins
        v.addFloat("marginBottom", 25.4, "Margin/Bottom", 0.0, 900.0)
        v.addFloat("marginLeft", 38.1, "Margin/Left", 0.0, 900.0)
        v.addFloat("marginRight", 25.4, "Margin/Right", 0.0, 900.0)
        v.addFloat("marginTop", 12.7, "Margin/Top", 0.0, 900.0)

        # paper size
        v.addFloat("paperHeight", 297.0, "Paper/Height", 100.0, 1000.0)
        v.addFloat("paperWidth", 210.0, "Paper/Width", 50.0, 1000.0)

        # leave at least this many action lines on the end of a page
        v.addInt("pbActionLines", 2, "PageBreakActionLines", 1, 30)

        # leave at least this many dialogue lines on the end of a page
        v.addInt("pbDialogueLines", 2, "PageBreakDialogueLines", 1, 30)

        # whether scene continueds are enabled
        v.addBool("sceneContinueds", False, "SceneContinueds")

        # scene continued text indent width
        v.addInt("sceneContinuedIndent", 45, "SceneContinuedIndent", -20, 80)

        # whether to include scene numbers
        v.addBool("pdfShowSceneNumbers", False, "ShowSceneNumbers")

        # whether to include PDF TOC
        v.addBool("pdfIncludeTOC", True, "IncludeTOC")

        # whether to show PDF TOC by default
        v.addBool("pdfShowTOC", True, "ShowTOC")

        # whether to open PDF document on current page
        v.addBool("pdfOpenOnCurrentPage", True, "OpenOnCurrentPage")

        # whether to remove Note elements in PDF output
        v.addBool("pdfRemoveNotes", False, "RemoveNotes")

        # whether to draw rectangles around the outlines of Note elements
        v.addBool("pdfOutlineNotes", True, "OutlineNotes")

        # whether to draw rectangle showing margins
        v.addBool("pdfShowMargins", False, "ShowMargins")

        # whether to show line numbers next to each line
        v.addBool("pdfShowLineNumbers", False, "ShowLineNumbers")

        # cursor position, line
        v.addInt("cursorLine", 0, "Cursor/Line", 0, 1000000)

        # cursor position, column
        v.addInt("cursorColumn", 0, "Cursor/Column", 0, 1000000)

        # various strings we add to the script
        v.addStrLatin1("strMore", "(MORE)", "String/MoreDialogue")
        v.addStrLatin1("strContinuedPageEnd", "(CONTINUED)", "String/ContinuedPageEnd")
        v.addStrLatin1(
            "strContinuedPageStart", "CONTINUED:", "String/ContinuedPageStart"
        )
        v.addStrLatin1("strDialogueContinued", " (cont'd)", "String/DialogueContinued")

        v.makeDicts()

    # load config from string 's'. does not throw any exceptions, silently
    # ignores any errors, and always leaves config in an ok state.
    def load(self, s):
        vals = self.cvars.makeVals(s)

        self.cvars.load(vals, "", self)

        for t in self.types.values():
            t.load(vals, "Element/")

        for pf in self.pdfFonts.values():
            pf.load(vals, "Font/")

        self.recalc()

    # save config into a string and return that.
    def save(self):
        s = self.cvars.save("", self)

        for t in self.types.values():
            s += t.save("Element/")

        for pf in self.pdfFonts.values():
            s += pf.save("Font/")

        return s

    # fix up all invalid config values and recalculate all variables
    # dependent on other variables.
    #
    # if doAll is False, enforces restrictions only on a per-variable
    # basis, e.g. doesn't modify variable v2 based on v1's value. this is
    # useful when user is interactively modifying v1, and it temporarily
    # strays out of bounds (e.g. when deleting the old text in an entry
    # box, thus getting the minimum value), which would then possibly
    # modify the value of other variables which is not what we want.
    def recalc(self, doAll=True):
        for it in self.cvars.numeric.values():
            util.clampObj(self, it.name, it.minVal, it.maxVal)

        for el in self.types.values():
            for it in el.cvars.numeric.values():
                util.clampObj(el, it.name, it.minVal, it.maxVal)

        for it in self.cvars.stringLatin1.values():
            setattr(self, it.name, util.toInputStr(getattr(self, it.name)))

        for pf in self.pdfFonts.values():
            pf.refresh()

        # make sure usable space on the page isn't too small
        if doAll and (self.marginTop + self.marginBottom) >= (self.paperHeight - 100.0):
            self.marginTop = 0.0
            self.marginBottom = 0.0

        h = self.paperHeight - self.marginTop - self.marginBottom

        # how many lines on a page
        self.linesOnPage = int(h / util.getTextHeight(self.fontSize))

    def getType(self, lt):
        return self.types[lt]

    # get a PDFFontInfo object for the given font type (PDF_FONT_*)
    def getPDFFont(self, fontType):
        return self.pdfFonts[fontType]

    # return a tuple of all the PDF font types
    def getPDFFontIds(self):
        return (PDF_FONT_NORMAL, PDF_FONT_BOLD, PDF_FONT_ITALIC, PDF_FONT_BOLD_ITALIC)


# global config. there is only ever one of these active.
class ConfigGlobal:
    cvars = None

    def __init__(self):

        if not self.__class__.cvars:
            self.setupVars()

        self.__class__.cvars.setDefaults(self)

        # type configs, key = line type, value = TypeGlobal
        self.types = {}

        # element types
        t = TypeGlobal(screenplay.SCENE)
        t.newTypeEnter = screenplay.ACTION
        t.newTypeTab = screenplay.CHARACTER
        t.nextTypeTab = screenplay.ACTION
        t.prevTypeTab = screenplay.TRANSITION
        self.types[t.lt] = t

        t = TypeGlobal(screenplay.ACTION)
        t.newTypeEnter = screenplay.ACTION
        t.newTypeTab = screenplay.CHARACTER
        t.nextTypeTab = screenplay.CHARACTER
        t.prevTypeTab = screenplay.CHARACTER
        self.types[t.lt] = t

        t = TypeGlobal(screenplay.CHARACTER)
        t.newTypeEnter = screenplay.DIALOGUE
        t.newTypeTab = screenplay.PAREN
        t.nextTypeTab = screenplay.ACTION
        t.prevTypeTab = screenplay.ACTION
        self.types[t.lt] = t

        t = TypeGlobal(screenplay.DIALOGUE)
        t.newTypeEnter = screenplay.CHARACTER
        t.newTypeTab = screenplay.ACTION
        t.nextTypeTab = screenplay.PAREN
        t.prevTypeTab = screenplay.ACTION
        self.types[t.lt] = t

        t = TypeGlobal(screenplay.PAREN)
        t.newTypeEnter = screenplay.DIALOGUE
        t.newTypeTab = screenplay.ACTION
        t.nextTypeTab = screenplay.CHARACTER
        t.prevTypeTab = screenplay.DIALOGUE
        self.types[t.lt] = t

        t = TypeGlobal(screenplay.TRANSITION)
        t.newTypeEnter = screenplay.SCENE
        t.newTypeTab = screenplay.TRANSITION
        t.nextTypeTab = screenplay.SCENE
        t.prevTypeTab = screenplay.CHARACTER
        self.types[t.lt] = t

        t = TypeGlobal(screenplay.SHOT)
        t.newTypeEnter = screenplay.ACTION
        t.newTypeTab = screenplay.CHARACTER
        t.nextTypeTab = screenplay.ACTION
        t.prevTypeTab = screenplay.SCENE
        self.types[t.lt] = t

        t = TypeGlobal(screenplay.ACTBREAK)
        t.newTypeEnter = screenplay.SCENE
        t.newTypeTab = screenplay.ACTION
        t.nextTypeTab = screenplay.SCENE
        t.prevTypeTab = screenplay.SCENE
        self.types[t.lt] = t

        t = TypeGlobal(screenplay.NOTE)
        t.newTypeEnter = screenplay.ACTION
        t.newTypeTab = screenplay.CHARACTER
        t.nextTypeTab = screenplay.ACTION
        t.prevTypeTab = screenplay.CHARACTER
        self.types[t.lt] = t

        # keyboard commands. these must be in alphabetical order.
        self.commands = (
            []
            if "TRELBY_TESTING" in os.environ
            else [
                Command(
                    "Abort",
                    "Abort something, e.g. selection," " auto-completion, etc.",
                    [wx.WXK_ESCAPE],
                    isFixed=True,
                ),
                Command("About", "Show the about dialog.", isMenu=True),
                Command(
                    "AutoCompletionDlg", "Open the auto-completion dialog.", isMenu=True
                ),
                Command(
                    "ChangeToActBreak",
                    "Change current element's style to" " act break.",
                    [util.Key(ord("B"), alt=True).toInt()],
                ),
                Command(
                    "ChangeToAction",
                    "Change current element's style to" " action.",
                    [util.Key(ord("A"), alt=True).toInt()],
                ),
                Command(
                    "ChangeToCharacter",
                    "Change current element's style to" " character.",
                    [util.Key(ord("C"), alt=True).toInt()],
                ),
                Command(
                    "ChangeToDialogue",
                    "Change current element's style to" " dialogue.",
                    [util.Key(ord("D"), alt=True).toInt()],
                ),
                Command(
                    "ChangeToNote",
                    "Change current element's style to note.",
                    [util.Key(ord("N"), alt=True).toInt()],
                ),
                Command(
                    "ChangeToParenthetical",
                    "Change current element's" " style to parenthetical.",
                    [util.Key(ord("P"), alt=True).toInt()],
                ),
                Command(
                    "ChangeToScene",
                    "Change current element's style to" " scene.",
                    [util.Key(ord("S"), alt=True).toInt()],
                ),
                Command("ChangeToShot", "Change current element's style to" " shot."),
                Command(
                    "ChangeToTransition",
                    "Change current element's style to" " transition.",
                    [util.Key(ord("T"), alt=True).toInt()],
                ),
                Command("CharacterMap", "Open the character map.", isMenu=True),
                Command(
                    "CloseScript",
                    "Close the current script.",
                    [util.Key(23, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command("CompareScripts", "Compare two scripts.", isMenu=True),
                Command(
                    "Copy",
                    "Copy selected text to the internal clipboard.",
                    [util.Key(3, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command(
                    "CopySystemCb",
                    "Copy selected text to the system's" " clipboard, unformatted.",
                    isMenu=True,
                ),
                Command(
                    "CopySystemCbFormatted",
                    "Copy selected text to the system's" " clipboard, formatted.",
                    isMenu=True,
                ),
                Command(
                    "Cut",
                    "Cut selected text to internal clipboard.",
                    [util.Key(24, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command(
                    "Delete",
                    "Delete the character under the cursor," " or selected text.",
                    [wx.WXK_DELETE],
                    isFixed=True,
                ),
                Command(
                    "DeleteBackward",
                    "Delete the character behind the" " cursor.",
                    [wx.WXK_BACK, util.Key(wx.WXK_BACK, shift=True).toInt()],
                    isFixed=True,
                ),
                Command(
                    "DeleteElements", "Open the 'Delete elements' dialog.", isMenu=True
                ),
                Command("ExportScript", "Export the current script.", isMenu=True),
                Command(
                    "FindAndReplaceDlg",
                    "Open the 'Find & Replace' dialog.",
                    [util.Key(6, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command(
                    "FindNextError",
                    "Find next error in the current script.",
                    [util.Key(5, ctrl=True).toInt()],
                    isMenu=True,
                ),
                Command(
                    "ForcedLineBreak",
                    "Insert a forced line break.",
                    [
                        util.Key(wx.WXK_RETURN, ctrl=True).toInt(),
                        util.Key(wx.WXK_RETURN, shift=True).toInt(),
                        # CTRL+Enter under wxMSW
                        util.Key(10, ctrl=True).toInt(),
                    ],
                    isFixed=True,
                ),
                Command(
                    "Fullscreen",
                    "Toggle fullscreen.",
                    [util.Key(wx.WXK_F11).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command(
                    "GotoPage",
                    "Goto to a given page.",
                    [util.Key(7, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command(
                    "GotoScene",
                    "Goto to a given scene.",
                    [util.Key(ord("G"), alt=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command("HeadersDlg", "Open the headers dialog.", isMenu=True),
                Command(
                    "HelpCommands",
                    "Show list of commands and their key" " bindings.",
                    isMenu=True,
                ),
                Command("HelpManual", "Open the manual.", isMenu=True),
                Command("ImportScript", "Import a script.", isMenu=True),
                Command(
                    "InsertNbsp",
                    "Insert non-breaking space.",
                    [util.Key(wx.WXK_SPACE, shift=True, ctrl=True).toInt()],
                    isMenu=True,
                ),
                Command(
                    "LoadScriptSettings", "Load script-specific settings.", isMenu=True
                ),
                Command("LoadSettings", "Load global settings.", isMenu=True),
                Command("LocationsDlg", "Open the locations dialog.", isMenu=True),
                Command(
                    "MoveDown",
                    "Move down.",
                    [wx.WXK_DOWN],
                    isMovement=True,
                    scrollDirection=SCROLL_DOWN,
                ),
                Command(
                    "MoveEndOfLine",
                    "Move to the end of the line or" " finish auto-completion.",
                    [wx.WXK_END],
                    isMovement=True,
                ),
                Command(
                    "MoveEndOfScript",
                    "Move to the end of the script.",
                    [util.Key(wx.WXK_END, ctrl=True).toInt()],
                    isMovement=True,
                ),
                Command("MoveLeft", "Move left.", [wx.WXK_LEFT], isMovement=True),
                Command(
                    "MoveNextWord",
                    "Move to start of next word.",
                    [util.Key(wx.WXK_RIGHT, ctrl=True).toInt()],
                    isMovement=True,
                    isFixed=True,
                ),
                Command(
                    "MovePageDown",
                    "Move one page down.",
                    [wx.WXK_PAGEDOWN],
                    isMovement=True,
                ),
                Command(
                    "MovePageUp", "Move one page up.", [wx.WXK_PAGEUP], isMovement=True
                ),
                Command(
                    "MovePrevWord",
                    "Move to start of previous word.",
                    [util.Key(wx.WXK_LEFT, ctrl=True).toInt()],
                    isMovement=True,
                    isFixed=True,
                ),
                Command("MoveRight", "Move right.", [wx.WXK_RIGHT], isMovement=True),
                Command(
                    "MoveSceneDown",
                    "Move one scene down.",
                    [util.Key(wx.WXK_DOWN, ctrl=True).toInt()],
                    isMovement=True,
                ),
                Command(
                    "MoveSceneUp",
                    "Move one scene up.",
                    [util.Key(wx.WXK_UP, ctrl=True).toInt()],
                    isMovement=True,
                ),
                Command(
                    "MoveStartOfLine",
                    "Move to the start of the line.",
                    [wx.WXK_HOME],
                    isMovement=True,
                ),
                Command(
                    "MoveStartOfScript",
                    "Move to the start of the" " script.",
                    [util.Key(wx.WXK_HOME, ctrl=True).toInt()],
                    isMovement=True,
                ),
                Command(
                    "MoveUp",
                    "Move up.",
                    [wx.WXK_UP],
                    isMovement=True,
                    scrollDirection=SCROLL_UP,
                ),
                Command(
                    "NameDatabase", "Open the character name database.", isMenu=True
                ),
                Command(
                    "NewElement", "Create a new element.", [wx.WXK_RETURN], isFixed=True
                ),
                Command(
                    "NewScript",
                    "Create a new script.",
                    [util.Key(14, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command(
                    "OpenScript",
                    "Open a script.",
                    [util.Key(15, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command("Paginate", "Paginate current script.", isMenu=True),
                Command(
                    "Paste",
                    "Paste text from the internal clipboard.",
                    [util.Key(22, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command(
                    "PasteSystemCb",
                    "Paste text from the system's" " clipboard.",
                    isMenu=True,
                ),
                Command(
                    "PrintScript",
                    "Print current script.",
                    [util.Key(16, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command(
                    "Quit",
                    "Quit the program.",
                    [util.Key(17, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command(
                    "Redo",
                    "Redo a change that was reverted through undo.",
                    [util.Key(25, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command("ReportCharacter", "Generate character report.", isMenu=True),
                Command(
                    "ReportDialogueChart",
                    "Generate dialogue chart report.",
                    isMenu=True,
                ),
                Command("ReportLocation", "Generate location report.", isMenu=True),
                Command("ReportScene", "Generate scene report.", isMenu=True),
                Command("ReportScript", "Generate script report.", isMenu=True),
                Command(
                    "RevertScript",
                    "Revert current script to the" " version on disk.",
                    isMenu=True,
                ),
                Command(
                    "SaveScript",
                    "Save the current script.",
                    [util.Key(19, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command(
                    "SaveScriptAs",
                    "Save the current script to a new file.",
                    isMenu=True,
                ),
                Command(
                    "SaveScriptSettingsAs",
                    "Save script-specific settings" " to a new file.",
                    isMenu=True,
                ),
                Command(
                    "SaveSettingsAs", "Save global settings to a new file.", isMenu=True
                ),
                Command(
                    "ScriptNext",
                    "Change to next open script.",
                    [
                        util.Key(wx.WXK_TAB, ctrl=True).toInt(),
                        util.Key(wx.WXK_PAGEDOWN, ctrl=True).toInt(),
                    ],
                    isMenu=True,
                ),
                Command(
                    "ScriptPrev",
                    "Change to previous open script.",
                    [
                        util.Key(wx.WXK_TAB, shift=True, ctrl=True).toInt(),
                        util.Key(wx.WXK_PAGEUP, ctrl=True).toInt(),
                    ],
                    isMenu=True,
                ),
                Command(
                    "ScriptSettings", "Change script-specific settings.", isMenu=True
                ),
                Command("SelectAll", "Select the entire script.", isMenu=True),
                Command(
                    "SelectScene",
                    "Select the current scene.",
                    [util.Key(1, ctrl=True).toInt()],
                    isMenu=True,
                ),
                Command(
                    "SetMark",
                    "Set mark at current cursor position.",
                    [util.Key(wx.WXK_SPACE, ctrl=True).toInt()],
                ),
                Command("Settings", "Change global settings.", isMenu=True),
                Command(
                    "SpellCheckerDictionaryDlg",
                    "Open the global spell checker dictionary dialog.",
                    isMenu=True,
                ),
                Command(
                    "SpellCheckerDlg",
                    "Spell check the script.",
                    [util.Key(wx.WXK_F8).toInt()],
                    isMenu=True,
                ),
                Command(
                    "SpellCheckerScriptDictionaryDlg",
                    "Open the script-specific spell checker dictionary" " dialog.",
                    isMenu=True,
                ),
                Command(
                    "Tab",
                    "Change current element to the next style or"
                    " create a new element.",
                    [wx.WXK_TAB],
                    isFixed=True,
                ),
                Command(
                    "TabPrev",
                    "Change current element to the previous" " style.",
                    [util.Key(wx.WXK_TAB, shift=True).toInt()],
                    isFixed=True,
                ),
                Command("TitlesDlg", "Open the titles dialog.", isMenu=True),
                Command(
                    "ToggleShowFormatting",
                    "Toggle 'Show formatting'" " display.",
                    isMenu=True,
                ),
                Command(
                    "Undo",
                    "Undo the last change.",
                    [util.Key(26, ctrl=True).toInt()],
                    isFixed=True,
                    isMenu=True,
                ),
                Command("ViewModeDraft", "Change view mode to draft.", isMenu=True),
                Command("ViewModeLayout", "Change view mode to layout.", isMenu=True),
                Command(
                    "ViewModeSideBySide",
                    "Change view mode to side by" " side.",
                    isMenu=True,
                ),
                Command("Watermark", "Generate watermarked PDFs.", isMenu=True),
            ]
        )

        self.recalc()

    def setupVars(self):
        v = self.__class__.cvars = mypickle.Vars()

        # how many seconds to show splash screen for on startup (0 = disabled)
        v.addInt("splashTime", 2, "SplashTime", 0, 10)

        # vertical distance between rows, in pixels
        v.addInt("fontYdelta", 18, "FontYDelta", 4, 125)

        # how many lines to scroll per mouse wheel event
        v.addInt("mouseWheelLines", 4, "MouseWheelLines", 1, 50)

        # interval in seconds between automatic pagination (0 = disabled)
        v.addInt("paginateInterval", 1, "PaginateInterval", 0, 10)

        # whether to check script for errors before export / print
        v.addBool("checkOnExport", True, "CheckScriptForErrors")

        # whether to auto-capitalize start of sentences
        v.addBool("capitalize", True, "CapitalizeSentences")

        # whether to auto-capitalize i -> I
        v.addBool("capitalizeI", True, "CapitalizeI")

        # whether to open scripts on their last saved position
        v.addBool("honorSavedPos", True, "OpenScriptOnSavedPos")

        # whether to recenter screen when cursor moves out of it
        v.addBool("recenterOnScroll", False, "RecenterOnScroll")

        # whether to overwrite selected text on typing
        v.addBool("overwriteSelectionOnInsert", True, "OverwriteSelectionOnInsert")

        # whether to use per-elem-type colors (textSceneColor etc.)
        # instead of using textColor for all elem types
        v.addBool("useCustomElemColors", False, "UseCustomElemColors")

        # page break indicators to show
        v.addInt("pbi", PBI_REAL, "PageBreakIndicators", PBI_FIRST, PBI_LAST)

        # PDF viewer program and args. defaults are empty since generating
        # them is a complex process handled by findPDFViewer.
        v.addStrUnicode("pdfViewerPath", "", "PDF/ViewerPath")
        v.addStrBinary("pdfViewerArgs", "", "PDF/ViewerArguments")

        # fonts. real defaults are set in setDefaultFonts.
        v.addStrBinary("fontNormal", "", "FontNormal")
        v.addStrBinary("fontBold", "", "FontBold")
        v.addStrBinary("fontItalic", "", "FontItalic")
        v.addStrBinary("fontBoldItalic", "", "FontBoldItalic")

        # default script directory
        v.addStrUnicode("scriptDir", misc.progPath, "DefaultScriptDirectory")

        # colors
        v.addColor("text", 0, 0, 0, "TextFG", "Text foreground")
        v.addColor(
            "textHdr", 128, 128, 128, "TextHeadersFG", "Text foreground (headers)"
        )
        v.addColor("textBg", 255, 255, 255, "TextBG", "Text background")
        v.addColor("workspace", 237, 237, 237, "Workspace", "Workspace")
        v.addColor("pageBorder", 202, 202, 202, "PageBorder", "Page border")
        v.addColor("pageShadow", 153, 153, 153, "PageShadow", "Page shadow")
        v.addColor("selected", 200, 200, 200, "Selected", "Selection")
        v.addColor("cursor", 135, 135, 253, "Cursor", "Cursor")
        v.addColor(
            "autoCompFg", 0, 0, 0, "AutoCompletionFG", "Auto-completion foreground"
        )
        v.addColor(
            "autoCompBg",
            255,
            240,
            168,
            "AutoCompletionBG",
            "Auto-completion background",
        )
        v.addColor("note", 255, 237, 223, "ScriptNote", "Script note")
        v.addColor("pagebreak", 221, 221, 221, "PageBreakLine", "Page-break line")
        v.addColor(
            "pagebreakNoAdjust",
            221,
            221,
            221,
            "PageBreakNoAdjustLine",
            "Page-break (original, not adjusted) line",
        )

        v.addColor("tabText", 50, 50, 50, "TabText", "Tab text")
        v.addColor("tabBorder", 202, 202, 202, "TabBorder", "Tab border")
        v.addColor("tabBarBg", 221, 217, 215, "TabBarBG", "Tab bar background")
        v.addColor("tabNonActiveBg", 180, 180, 180, "TabNonActiveBg", "Tab, non-active")

        for t in getTIs():
            v.addColor(
                "text%s" % t.name,
                0,
                0,
                0,
                "Text%sFG" % t.name,
                "Text foreground for %s" % t.name,
            )

        v.makeDicts()

    # load config from string 's'. does not throw any exceptions, silently
    # ignores any errors, and always leaves config in an ok state.
    def load(self, s):
        vals = self.cvars.makeVals(s)

        self.cvars.load(vals, "", self)

        for t in self.types.values():
            t.load(vals, "Element/")

        for cmd in self.commands:
            cmd.load(vals, "Command/")

        self.recalc()

    # save config into a string and return that.
    def save(self):
        s = self.cvars.save("", self)

        for t in self.types.values():
            s += t.save("Element/")

        for cmd in self.commands:
            s += cmd.save("Command/")

        return s

    # fix up all invalid config values.
    def recalc(self):
        for it in self.cvars.numeric.values():
            util.clampObj(self, it.name, it.minVal, it.maxVal)

    def getType(self, lt):
        return self.types[lt]

    # add SHIFT+Key alias for all keys bound to movement commands, so
    # selection-movement works.
    def addShiftKeys(self):
        for cmd in self.commands:
            if cmd.isMovement:
                nk = []

                for key in cmd.keys:
                    k = util.Key.fromInt(key)
                    k.shift = True
                    ki = k.toInt()

                    if ki not in cmd.keys:
                        nk.append(ki)

                cmd.keys.extend(nk)

    # remove key (int) from given cmd
    def removeKey(self, cmd, key):
        cmd.keys.remove(key)

        if cmd.isMovement:
            k = util.Key.fromInt(key)
            k.shift = True
            ki = k.toInt()

            if ki in cmd.keys:
                cmd.keys.remove(ki)

    # get textual description of conflicting keys, or None if no
    # conflicts.
    def getConflictingKeys(self):
        keys = {}

        for cmd in self.commands:
            for key in cmd.keys:
                if key in keys:
                    keys[key].append(cmd.name)
                else:
                    keys[key] = [cmd.name]

        s = ""
        for k, v in keys.items():
            if len(v) > 1:
                s += "%s:" % util.Key.fromInt(k).toStr()

                for cmd in v:
                    s += " %s" % cmd

                s += "\n"

        if s == "":
            return None
        else:
            return s

    # set default values that vary depending on platform, wxWidgets
    # version, etc. this is not at the end of __init__ because
    # non-interactive uses have no needs for these.
    def setDefaults(self):
        # check keyboard commands are listed in correct order
        commands = [cmd.name for cmd in self.commands]
        commandsSorted = sorted(commands)

        if commands != commandsSorted:
            # for i in range(len(commands)):
            #     if commands[i] != commandsSorted[i]:
            #         print "Got: %s Expected: %s" % (commands[i], commandsSorted[i])

            # if you get this error, you've put a new command you've added
            # in an incorrect place in the command list. uncomment the
            # above lines to figure out where it should be.
            raise ConfigError("Commands not listed in correct order")

        self.setDefaultFonts()
        self.findPDFViewer()

    # set default fonts
    def setDefaultFonts(self):
        fn = ["", "", "", ""]

        if misc.isUnix:
            fn[0] = "Monospace 12"
            fn[1] = "Monospace Bold 12"
            fn[2] = "Monospace Italic 12"
            fn[3] = "Monospace Bold Italic 12"

        elif misc.isWindows:
            fn[0] = "0;-13;0;0;0;400;0;0;0;0;3;2;1;49;Courier New"
            fn[1] = "0;-13;0;0;0;700;0;0;0;0;3;2;1;49;Courier New"
            fn[2] = "0;-13;0;0;0;400;255;0;0;0;3;2;1;49;Courier New"
            fn[3] = "0;-13;0;0;0;700;255;0;0;0;3;2;1;49;Courier New"

        else:
            raise ConfigError("Unknown platform")

        self.fontNormal = fn[0]
        self.fontBold = fn[1]
        self.fontItalic = fn[2]
        self.fontBoldItalic = fn[3]

    # set PDF viewer program to the best one found on the machine.
    def findPDFViewer(self):
        name, args = util.getPDFViewer()
        if name:
            self.pdfViewerPath = name
            self.pdfViewerArgs = args


# config stuff that are wxwindows objects, so can't be in normal
# ConfigGlobal (deepcopy dies)
class ConfigGui:

    # constants
    constantsInited = False
    bluePen = None
    redColor = None
    blackColor = None

    def __init__(self, cfgGl):

        if not ConfigGui.constantsInited:
            ConfigGui.bluePen = wx.Pen(wx.Colour(0, 0, 255))
            ConfigGui.redColor = wx.Colour(255, 0, 0)
            ConfigGui.blackColor = wx.Colour(0, 0, 0)

            ConfigGui.constantsInited = True

        # convert cfgGl.MyColor -> cfgGui.wx.Colour
        for it in cfgGl.cvars.color.values():
            c = getattr(cfgGl, it.name)
            tmp = wx.Colour(c.r, c.g, c.b)
            setattr(self, it.name, tmp)

        # key = line type, value = wx.Colour
        self._lt2textColor = {}

        for t in getTIs():
            self._lt2textColor[t.lt] = getattr(self, "text%sColor" % t.name)

        self.textPen = wx.Pen(self.textColor)
        self.textHdrPen = wx.Pen(self.textHdrColor)

        self.workspaceBrush = wx.Brush(self.workspaceColor)
        self.workspacePen = wx.Pen(self.workspaceColor)

        self.textBgBrush = wx.Brush(self.textBgColor)
        self.textBgPen = wx.Pen(self.textBgColor)

        self.pageBorderPen = wx.Pen(self.pageBorderColor)
        self.pageShadowPen = wx.Pen(self.pageShadowColor)

        self.selectedBrush = wx.Brush(self.selectedColor)
        self.selectedPen = wx.Pen(self.selectedColor)

        self.cursorBrush = wx.Brush(self.cursorColor)
        self.cursorPen = wx.Pen(self.cursorColor)

        self.noteBrush = wx.Brush(self.noteColor)
        self.notePen = wx.Pen(self.noteColor)

        self.autoCompPen = wx.Pen(self.autoCompFgColor)
        self.autoCompBrush = wx.Brush(self.autoCompBgColor)
        self.autoCompRevPen = wx.Pen(self.autoCompBgColor)
        self.autoCompRevBrush = wx.Brush(self.autoCompFgColor)

        self.pagebreakPen = wx.Pen(self.pagebreakColor)
        self.pagebreakNoAdjustPen = wx.Pen(self.pagebreakNoAdjustColor, style=wx.DOT)

        self.tabTextPen = wx.Pen(self.tabTextColor)
        self.tabBorderPen = wx.Pen(self.tabBorderColor)

        self.tabBarBgBrush = wx.Brush(self.tabBarBgColor)
        self.tabBarBgPen = wx.Pen(self.tabBarBgColor)

        self.tabNonActiveBgBrush = wx.Brush(self.tabNonActiveBgColor)
        self.tabNonActiveBgPen = wx.Pen(self.tabNonActiveBgColor)

        # a 4-item list of FontInfo objects, indexed by the two lowest
        # bits of pml.TextOp.flags.
        self.fonts = []

        for fname in ["fontNormal", "fontBold", "fontItalic", "fontBoldItalic"]:
            fi = FontInfo()

            s = getattr(cfgGl, fname)

            # evil users can set the font name to empty by modifying the
            # config file, and some wxWidgets ports crash hard when trying
            # to create a font from an empty string, so we must guard
            # against that.
            if s:
                nfi = wx.NativeFontInfo()
                nfi.FromString(s)

                try:
                    fi.font = wx.Font(nfi)

                    # likewise, evil users can set the font name to "z" or
                    # something equally silly, resulting in an
                    # invalid/non-existent font. on wxGTK2 and wxMSW we can
                    # detect this by checking the point size of the font.
                    if fi.font.GetPointSize() == 0:
                        fi.font = None
                except wx._core.wxAssertionError:
                    # Some platforms (mac) will assert internally if point size
                    # is 0, preventing us from even trying to check, so just
                    # catch that too.
                    fi.font = None

            # if either of the above failures happened, create a dummy
            # font and use it. this sucks but is preferable to crashing or
            # displaying an empty screen.
            if not fi.font:
                fi.font = wx.Font(
                    10,
                    wx.MODERN,
                    wx.NORMAL,
                    wx.NORMAL,
                    encoding=wx.FONTENCODING_ISO8859_1,
                )
                setattr(cfgGl, fname, fi.font.GetNativeFontInfo().ToString())

            fx, fy = util.getTextExtent(fi.font, "O")

            fi.fx = max(1, fx)
            fi.fy = max(1, fy)

            self.fonts.append(fi)

    # TextType -> FontInfo
    def tt2fi(self, tt):
        return self.fonts[tt.isBold | (tt.isItalic << 1)]

    # line type -> wx.Colour
    def lt2textColor(self, lt):
        return self._lt2textColor[lt]


def _conv(dict, key, raiseException=True):
    val = dict.get(key)
    if (val == None) and raiseException:
        raise ConfigError("key '%s' not found from '%s'" % (key, dict))

    return val


# get TypeInfos
def getTIs():
    return _ti


def char2lb(char, raiseException=True):
    return _conv(_char2lb, char, raiseException)


def lb2char(lb):
    return _conv(_lb2char, lb)


def lb2str(lb):
    return _conv(_lb2str, lb)


def char2lt(char, raiseException=True):
    ti = _conv(_char2ti, char, raiseException)

    if ti:
        return ti.lt
    else:
        return None


def lt2char(lt):
    return _conv(_lt2ti, lt).char


def name2ti(name, raiseException=True):
    return _conv(_name2ti, name, raiseException)


def lt2ti(lt):
    return _conv(_lt2ti, lt)


def _init():

    for lt, char, name in (
        (screenplay.SCENE, "\\", "Scene"),
        (screenplay.ACTION, ".", "Action"),
        (screenplay.CHARACTER, "_", "Character"),
        (screenplay.DIALOGUE, ":", "Dialogue"),
        (screenplay.PAREN, "(", "Parenthetical"),
        (screenplay.TRANSITION, "/", "Transition"),
        (screenplay.SHOT, "=", "Shot"),
        (screenplay.ACTBREAK, "@", "Act break"),
        (screenplay.NOTE, "%", "Note"),
    ):

        ti = TypeInfo(lt, char, name)

        _ti.append(ti)
        _lt2ti[lt] = ti
        _char2ti[char] = ti
        _name2ti[name] = ti


_init()



import wx

import trelby.screenplay as screenplay

ID_EDIT_UNDO = wx.NewId()
ID_EDIT_REDO = wx.NewId()
ID_EDIT_COPY = wx.NewId()
ID_EDIT_COPY_SYSTEM = wx.NewId()
ID_EDIT_COPY_TO_CB = wx.NewId()
ID_EDIT_COPY_TO_CB_FMT = wx.NewId()
ID_EDIT_CUT = wx.NewId()
ID_EDIT_DELETE_ELEMENTS = wx.NewId()
ID_EDIT_FIND = wx.NewId()
ID_EDIT_GOTO_SCENE = wx.NewId()
ID_EDIT_GOTO_PAGE = wx.NewId()
ID_EDIT_INSERT_NBSP = wx.NewId()
ID_EDIT_PASTE = wx.NewId()
ID_EDIT_PASTE_FROM_CB = wx.NewId()
ID_EDIT_SELECT_ALL = wx.NewId()
ID_EDIT_SELECT_SCENE = wx.NewId()
ID_FILE_CLOSE = wx.NewId()
ID_FILE_EXIT = wx.NewId()
ID_FILE_EXPORT = wx.NewId()
ID_FILE_IMPORT = wx.NewId()
ID_FILE_NEW = wx.NewId()
ID_FILE_OPEN = wx.NewId()
ID_FILE_PRINT = wx.NewId()
ID_FILE_REVERT = wx.NewId()
ID_FILE_SAVE = wx.NewId()
ID_FILE_SAVE_AS = wx.NewId()
ID_FILE_SETTINGS = wx.NewId()
ID_HELP_ABOUT = wx.NewId()
ID_HELP_COMMANDS = wx.NewId()
ID_HELP_MANUAL = wx.NewId()
ID_REPORTS_CHARACTER_REP = wx.NewId()
ID_REPORTS_DIALOGUE_CHART = wx.NewId()
ID_REPORTS_LOCATION_REP = wx.NewId()
ID_REPORTS_SCENE_REP = wx.NewId()
ID_REPORTS_SCRIPT_REP = wx.NewId()
ID_SCRIPT_AUTO_COMPLETION = wx.NewId()
ID_SCRIPT_FIND_ERROR = wx.NewId()
ID_SCRIPT_HEADERS = wx.NewId()
ID_SCRIPT_LOCATIONS = wx.NewId()
ID_SCRIPT_PAGINATE = wx.NewId()
ID_SCRIPT_SC_DICT = wx.NewId()
ID_SCRIPT_SETTINGS = wx.NewId()
ID_SCRIPT_SETTINGS_CHANGE = wx.NewId()
ID_SCRIPT_SETTINGS_LOAD = wx.NewId()
ID_SCRIPT_SETTINGS_SAVE_AS = wx.NewId()
ID_SCRIPT_TITLES = wx.NewId()
ID_SETTINGS_CHANGE = wx.NewId()
ID_SETTINGS_LOAD = wx.NewId()
ID_SETTINGS_SAVE_AS = wx.NewId()
ID_SETTINGS_SC_DICT = wx.NewId()
ID_SHOW_HIDE_TOOLBAR = wx.NewId()
ID_TOOLS_CHARMAP = wx.NewId()
ID_TOOLS_COMPARE_SCRIPTS = wx.NewId()
ID_TOOLS_NAME_DB = wx.NewId()
ID_TOOLS_SPELL_CHECK = wx.NewId()
ID_TOOLS_WATERMARK = wx.NewId()
ID_VIEW_SHOW_FORMATTING = wx.NewId()
ID_VIEW_STYLE_DRAFT = wx.NewId()
ID_VIEW_STYLE_LAYOUT = wx.NewId()
ID_VIEW_STYLE_SIDE_BY_SIDE = wx.NewId()
ID_TOOLBAR_SETTINGS = wx.NewId()
ID_TOOLBAR_SCRIPTSETTINGS = wx.NewId()
ID_TOOLBAR_REPORTS = wx.NewId()
ID_TOOLBAR_VIEWS = wx.NewId()
ID_TOOLBAR_TOOLS = wx.NewId()
ID_VIEW_FULL_SCREEN = wx.NewId()
ID_ELEM_TO_ACTION = wx.NewId()
ID_ELEM_TO_CHARACTER = wx.NewId()
ID_ELEM_TO_DIALOGUE = wx.NewId()
ID_ELEM_TO_NOTE = wx.NewId()
ID_ELEM_TO_PAREN = wx.NewId()
ID_ELEM_TO_SCENE = wx.NewId()
ID_ELEM_TO_SHOT = wx.NewId()
ID_ELEM_TO_ACTBREAK = wx.NewId()
ID_ELEM_TO_TRANSITION = wx.NewId()

idToLTMap = {
    ID_ELEM_TO_SCENE: screenplay.SCENE,
    ID_ELEM_TO_ACTION: screenplay.ACTION,
    ID_ELEM_TO_CHARACTER: screenplay.CHARACTER,
    ID_ELEM_TO_DIALOGUE: screenplay.DIALOGUE,
    ID_ELEM_TO_PAREN: screenplay.PAREN,
    ID_ELEM_TO_TRANSITION: screenplay.TRANSITION,
    ID_ELEM_TO_SHOT: screenplay.SHOT,
    ID_ELEM_TO_ACTBREAK: screenplay.ACTBREAK,
    ID_ELEM_TO_NOTE: screenplay.NOTE,
}



import tests.u as u
import trelby.screenplay as scr

# tests adding characters, i.e. normal typing


def testSpaceAtEOL():
    sp = u.load()
    sp.cmd("moveDown", count=3)
    sp.cmd("moveLineEnd")
    sp.cmd("addChar", char="z")
    sp.cmd("addChar", char=" ")
    assert sp.lines[3].text.endswith("z ")
    sp.cmd("addChar", char="x")
    assert (sp.line == 4) and (sp.column == 1)
    assert sp.lines[3].lb == scr.LB_SPACE
    assert sp.lines[4].lb == scr.LB_LAST
    assert sp.lines[3].text.endswith("wouldz")
    assert sp.lines[4].text.startswith("x be")


def testNbspAtEOL():
    sp = u.load()
    sp.cmd("moveDown", count=3)
    sp.cmd("moveLineEnd")
    sp.cmd("addChar", char=chr(160))
    sp.cmd("addChar", char="a")
    assert sp.lines[3].text.endswith("mind")
    assert sp.lines[4].text.startswith("would%sa" % chr(160))
    assert (sp.line == 4) and (sp.column == 7)
    assert sp.lines[3].lb == scr.LB_SPACE
    assert sp.lines[4].lb == scr.LB_LAST


# FIXME: lot more tests



import uuid
from typing import AnyStr, Dict, Optional, Tuple

from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.pdfgen.canvas import Canvas

import trelby.pml as pml
import trelby.pml as util

# PDF transform matrixes where key is the angle from x-axis
# in counter-clockwise direction.
TRANSFORM_MATRIX = {
    45: (1, 1, -1, 1),
    90: (0, 1, -1, 0),
}


# users should only use this.
def generate(doc: "pml.Document") -> bytes:
    tmp = PDFExporter(doc)
    return tmp.generate()


# An abstract base class for all PDF drawing operations.
class PDFDrawOp:

    # write PDF drawing operations corresponding to the PML object pmlOp
    # to output (util.String). pe = PDFExporter.
    def draw(
        self, pmlOp: "pml.DrawOp", pageNr: int, pe: "PDFExporter", canvas: Canvas
    ) -> None:
        raise Exception("draw not implemented")


class PDFTextOp(PDFDrawOp):
    def draw(self, pmlOp: "pml.DrawOp", pageNr: int, pe: "PDFExporter", canvas) -> None:
        if not isinstance(pmlOp, pml.TextOp):
            raise Exception(
                "PDFTextOp is only compatible with pml.TextOp, got "
                + type(pmlOp).__name__
            )

        # we need to adjust y position since PDF uses baseline of text as
        # the y pos, but pml uses top of the text as y pos. The Adobe
        # standard Courier family font metrics give 157 units in 1/1000
        # point units as the Descender value, thus giving (1000 - 157) =
        # 843 units from baseline to top of text.

        # http://partners.adobe.com/asn/tech/type/ftechnotes.jsp contains
        # the "Font Metrics for PDF Core 14 Fonts" document.

        x = pe.x(pmlOp.x)
        y = pe.y(pmlOp.y) - 0.843 * pmlOp.size

        newFont = pe.getFontForFlags(pmlOp.flags)
        canvas.setFont(newFont, pmlOp.size)

        if pmlOp.angle is not None:
            matrix = TRANSFORM_MATRIX.get(pmlOp.angle)

            if matrix:
                canvas.addLiteral(
                    "BT\n"
                    "%f %f %f %f %f %f Tm\n"
                    "(%s) Tj\n"
                    "ET\n"
                    % (
                        matrix[0],
                        matrix[1],
                        matrix[2],
                        matrix[3],
                        x,
                        y,
                        pe.escapeStr(pmlOp.text),
                    )
                )  # TODO: Doing this with addLiteral, non-latin characters won't work in watermarks. There must be a better way to do this with reportlab, that we do it this way is for historical reasons and because no one took the time to change it
            else:
                # unsupported angle, don't print it.
                pass
        else:
            canvas.drawString(x, y, pmlOp.text)

        if pmlOp.flags & pml.UNDERLINED:

            undLen = canvas.stringWidth(pmlOp.text, newFont, pmlOp.size)

            # all standard PDF fonts have the underline line 100 units
            # below baseline with a thickness of 50
            undY = y - 0.1 * pmlOp.size
            canvas.setLineWidth(0.05 * pmlOp.size)

            canvas.line(x, undY, x + undLen, undY)

        # create bookmark for table of contents if applicable
        if pmlOp.toc:
            bookmarkKey = (
                uuid.uuid4().hex
            )  # we need a unique key to link the bookmark in toc – TODO: generate a more speaking one
            canvas.bookmarkHorizontal(bookmarkKey, pe.x(pmlOp.x), pe.y(pmlOp.y))
            canvas.addOutlineEntry(pmlOp.toc.text, bookmarkKey)


class PDFLineOp(PDFDrawOp):
    def draw(self, pmlOp: "pml.DrawOp", pageNr: int, pe: "PDFExporter", canvas):
        if not isinstance(pmlOp, pml.LineOp):
            raise Exception(
                "PDFLineOp is only compatible with pml.LineOp, got "
                + type(pmlOp).__name__
            )

        points = pmlOp.points
        numberOfPoints = len(points)

        if numberOfPoints < 2:
            print("LineOp contains only %d points" % numberOfPoints)

            return

        canvas.setLineWidth(pe.mm2points(pmlOp.width))

        lines = []
        for i in range(0, numberOfPoints - 1):
            lines.append(pe.xy(points[i]) + pe.xy(points[i + 1]))

        if pmlOp.isClosed:
            lines.append(pe.xy(points[i + 1]) + pe.xy(points[0]))

        canvas.lines(lines)


class PDFRectOp(PDFDrawOp):
    def draw(self, pmlOp: "pml.DrawOp", pageNr: int, pe: "PDFExporter", canvas) -> None:
        if not isinstance(pmlOp, pml.RectOp):
            raise Exception(
                "PDFRectOp is only compatible with pml.RectOp, got "
                + type(pmlOp).__name__
            )

        if pmlOp.lw != -1:
            canvas.setLineWidth(pe.mm2points(pmlOp.lw))

        canvas.rect(
            pe.x(pmlOp.x),
            pe.y(pmlOp.y) - pe.mm2points(pmlOp.height),
            pe.mm2points(pmlOp.width),
            pe.mm2points(pmlOp.height),
            pmlOp.fillType == pml.NO_FILL or pmlOp.fillType == pml.STROKE_FILL,
            pmlOp.fillType == pml.FILL or pmlOp.fillType == pml.STROKE_FILL,
        )


class PDFQuarterCircleOp(PDFDrawOp):
    def draw(self, pmlOp: "pml.DrawOp", pageNr: int, pe: "PDFExporter", canvas) -> None:
        if not isinstance(pmlOp, pml.QuarterCircleOp):
            raise Exception(
                "PDFQuarterCircleOp is only compatible with pml.QuarterCircleOp, got "
                + type(pmlOp).__name__
            )

        sX = pmlOp.flipX and -1 or 1
        sY = pmlOp.flipY and -1 or 1

        # The traditional constant is: 0.552284749
        # however, as described here:
        # http://spencermortensen.com/articles/bezier-circle/,
        # this has a maximum radial drift of 0.027253%.
        # The constant calculated by Spencer Mortensen
        # has a max. drift of 0.019608% which is 28% better.
        A = pmlOp.radius * 0.551915024494

        canvas.setLineWidth(pe.mm2points(pmlOp.width))
        canvas.bezier(
            pe.x(pmlOp.x - pmlOp.radius * sX),
            pe.y(pmlOp.y),
            pe.x(pmlOp.x - pmlOp.radius * sX),
            pe.y(pmlOp.y - A * sY),
            pe.x(pmlOp.x - A * sX),
            pe.y(pmlOp.y - pmlOp.radius * sY),
            pe.x(pmlOp.x),
            pe.y(pmlOp.y - pmlOp.radius * sY),
        )


class PDFArbitraryOp(PDFDrawOp):
    def draw(self, pmlOp: "pml.DrawOp", pageNr: int, pe: "PDFExporter", canvas) -> None:
        if not isinstance(pmlOp, pml.PDFOp):
            raise Exception(
                "PDFArbitraryOp is only compatible with pml.PDFOp, got "
                + type(pmlOp).__name__
            )

        canvas.addLiteral("%s\n" % pmlOp.cmds)


# used for keeping track of used fonts
class FontInfo:
    def __init__(self, name: str):
        self.name: str = name

        # font number (the name in the /F PDF command), or -1 if not used
        self.number: int = -1

        # PDFObject that contains the /Font object for this font, or None
        self.pdfObj: Optional[PDFObject] = None


# one object in a PDF file
class PDFObject:
    def __init__(self, nr: int, data: str = ""):
        # PDF object number
        self.nr: int = nr

        # all data between 'obj/endobj' tags, excluding newlines
        self.data: str = data

    # write object to canvas.
    def write(self, canvas: Canvas) -> None:
        code = "%d 0 obj\n" % self.nr
        code += self.data
        code += "\nendobj\n"
        canvas.addLiteral(code)


class PDFExporter:
    # see genWidths
    _widthsStr: Optional[str] = None

    def __init__(self, doc: "pml.Document"):
        self.doc: pml.Document = doc
        # fast lookup of font information
        self.fonts: Dict[int, FontInfo] = {
            pml.COURIER: FontInfo("Courier"),
            pml.COURIER | pml.BOLD: FontInfo("Courier-Bold"),
            pml.COURIER | pml.ITALIC: FontInfo("Courier-Oblique"),
            pml.COURIER | pml.BOLD | pml.ITALIC: FontInfo("Courier-BoldOblique"),
            pml.HELVETICA: FontInfo("Helvetica"),
            pml.HELVETICA | pml.BOLD: FontInfo("Helvetica-Bold"),
            pml.HELVETICA | pml.ITALIC: FontInfo("Helvetica-Oblique"),
            pml.HELVETICA | pml.BOLD | pml.ITALIC: FontInfo("Helvetica-BoldOblique"),
            pml.TIMES_ROMAN: FontInfo("Times-Roman"),
            pml.TIMES_ROMAN | pml.BOLD: FontInfo("Times-Bold"),
            pml.TIMES_ROMAN | pml.ITALIC: FontInfo("Times-Italic"),
            pml.TIMES_ROMAN | pml.BOLD | pml.ITALIC: FontInfo("Times-BoldItalic"),
        }

    # generate PDF document and return it as a string
    def generate(self) -> bytes:
        doc = self.doc
        canvas = Canvas(
            "",
            pdfVersion=(1, 5),
            pagesize=(self.mm2points(doc.w), self.mm2points(doc.h)),
            initialFontName=self.getFontForFlags(pml.NORMAL),
        )

        # set PDF info
        version = self.doc.version
        canvas.setCreator("Trelby " + version)
        canvas.setProducer("Trelby " + version)
        if self.doc.uniqueId:
            canvas.setKeywords(self.doc.uniqueId)

        if doc.defPage != -1:
            # canvas.addLiteral("/OpenAction [%d 0 R /XYZ null null 0]\n" % (self.pageObjs[0].nr + doc.defPage * 2)) # this should make the PDF reader open the PDF at the desired page
            # TODO: This doesn't seem to be easily doable with reportlab. /OpenAction is considered a security threat by some (as it allows executing JavaScript), so I think it's unlikely they'll add support. Also, this feature didn't work with many PDF viewers anyway; I tested Evince, Okular and pdf.js in Firefox, and they all didn't support it. So maybe, we should remove this feature entirely?
            pass

        numberOfPages: int = len(doc.pages)

        # draw pages
        for i in range(numberOfPages):
            pg = self.doc.pages[i]
            for op in pg.ops:
                op.pdfOp.draw(op, i, self, canvas)

            if i < numberOfPages - 1:
                canvas.showPage()

        if doc.showTOC:
            canvas.showOutline()

        return canvas.getpdfdata()

    # generate a stream object's contents. 's' is all data between
    # 'stream/endstream' tags, excluding newlines.
    def genStream(self, s, isFontStream=False) -> str:
        compress = False

        # embedded TrueType font program streams for some reason need a
        # Length1 entry that records the uncompressed length of the stream
        if isFontStream:
            lenStr = "/Length1 %d\n" % len(s)
        else:
            lenStr = ""

        filterStr = " "
        if compress:
            s = s.encode("zlib")
            filterStr = "/Filter /FlateDecode\n"

        return (
            "<< /Length %d\n%s%s>>\n"
            "stream\n"
            "%s\n"
            "endstream" % (len(s), lenStr, filterStr, s)
        )

    # get font name to use for given flags. also registers the font in reportlabs if it does not yet exist.
    def getFontForFlags(self, flags: int) -> str:
        # the "& 15" gets rid of the underline flag
        fontInfo = self.fonts.get(flags & 15)

        if not fontInfo:
            raise Exception("PDF.getfontNr: invalid flags %d" % flags)

        # the "& 15" gets rid of the underline flag
        customFontInfo = self.doc.fonts.get(flags & 15)

        if not customFontInfo:
            return fontInfo.name

        if not customFontInfo.name in pdfmetrics.getRegisteredFontNames():
            if not customFontInfo.fontFileName:
                raise Exception(
                    'Font name "%s" is not known and no font file name provided. Please provide a file name for this font in the settings or use the default font.'
                    % customFontInfo.name
                )
            pdfmetrics.registerFont(
                TTFont(customFontInfo.name, customFontInfo.fontFileName)
            )

        return customFontInfo.name

    # escape string
    def escapeStr(self, s: str) -> str:
        return s.replace("\\", "\\\\").replace("(", "\\(").replace(")", "\\)")

    # convert mm to points (1/72 inch).
    def mm2points(self, mm: float) -> float:
        # 2.834 = 72 / 25.4
        return mm * 2.83464567

    # convert x coordinate
    def x(self, x: float) -> float:
        return self.mm2points(x)

    # convert y coordinate
    def y(self, y: float) -> float:
        return self.mm2points(self.doc.h - y)

    # convert xy, which is (x, y) pair, into PDF coordinates
    def xy(self, xy: Tuple[float, float]) -> Tuple[float, float]:
        x = self.x(xy[0])
        y = self.y(xy[1])

        return (x, y)
